function  write_geometry(gname,boundary,range,fname,radius,ps,nc)
% Take a geometry structure and write a data file
% Now construct the index matrix for these elements
% boundary structure elements are
% .type   type of boundary, plane, cylinder, etc
% .name   description string 
% .min    minimum coordinates
% .max    maximum coordinates
% .v(3,:) vector parameters
% .s(1,:) scalar parameters

if nargin < 7
  nc = [1 1 1]; % Default is cuts along all axis
end
if nargin < 6
  ps = '%.15g'; %String to specify the precision used in writing the data 
end
if nargin < 5
  radius = -1; % What radius was the geometry constructed for
end
if nargin < 4
  fname = 'test.geom';
end
if nargin < 3
  range = [];    % No range
end

r_tol = 0.001; % What is the accuracy of distance measurements

if length(ps)==0
  ps = '%.15g'; %String to specify the precision used in writing the data 
end
if length(findstr('/',line))==0         % If no directory is specified
  fname = ['/home/jim/dem/data/' fname];
end

disp(['Writing geometry file ' fname]);

n = length(boundary);
for i = 1:n;
  ix(i,1) = boundary(i).min(1);
  ix(i,2) = boundary(i).max(1);
  iy(i,1) = boundary(i).min(2);
  iy(i,2) = boundary(i).max(2);
  iz(i,1) = boundary(i).min(3);
  iz(i,2) = boundary(i).max(3);
end
% Sort and remove duplicates
ix = unduplicate(ix(:),r_tol);
iy = unduplicate(iy(:),r_tol);
iz = unduplicate(iz(:),r_tol);
ix(isinf(ix))=[];
iy(isinf(iy))=[];
iz(isinf(iz))=[];

if ~nc(1)
  ix = []; % No x cuts;
end
if ~nc(2)
  iy = []; % No y cuts;
end
if ~nc(3)
  iz = []; % No z cuts;
end

gdim = [length(ix) length(iy) length(iz)]; % Number of cuts in each dimension
% Now step through the boundaries and hash them onto the grid
index_c=zeros(gdim+1);
index_d=zeros([gdim+1 5]);
for i = 1:n;
  %Find range of boxes spanned by region reduced by tolerance
  bmin = boundary(i).min+r_tol;
  bmax = boundary(i).max-r_tol;
% $$$   rg = 100*eps*(bmax-bmin); 
% $$$   if (isinf(rg))
% $$$     rg = eps*sign(rg);
% $$$   end
% $$$   bmin = bmin + rg;
% $$$   bmax = bmax - rg; 
  if any(bmin>bmax)
    error('Boundary is impossibly thin');
  end
  c1 = 1+hash3(bmin,ix,iy,iz);
  c2 = 1+hash3(bmax,ix,iy,iz);
  if any(c2<c1) error('indexes out of order');end
  [index_c,index_d]=add_to_index(index_c,index_d,c1,c2,i);
 t = find(index_d==i);
end

%Now write the data to the file
% First write out the name list
fid = fopen(fname,'w');
fprintf(fid,[' &GEOMETRY_PARAM\n']);
fprintf(fid,['     GEOM_NAME   = "' gname '"\n']);
fprintf(fid,['     GEOM_TYPE   = "general"\n']);
fprintf(fid,['     GEOM_RADIUS = ' ps '\n'], radius);
fprintf(fid,['     GEOM_NUM    = %i'],n);
if length(range)==6
  fprintf(fid,'\n     GEOM_BOUNDS =');
  fprintf(fid,[' ' ps],range);
end
  fprintf(fid,'/\n\n');
for i=1:n  
  structfprintf(fid,boundary(i),i,ps);
end

% Write out cut points
fprintf(fid,' %i', gdim);fprintf(fid,',\n'); 
fprintf(fid,[' ' ps], ix);fprintf(fid,',\n'); 
fprintf(fid,[' ' ps], iy);fprintf(fid,',\n'); 
fprintf(fid,[' ' ps], iz);fprintf(fid,',\n\n'); 

% write out count
fprintf(fid,[' ' ps],index_c);
fprintf(fid,',\n\n');

% write out boundary numbers
for i=1:max(index_c(:))
  fprintf(fid,[' ' ps],index_d(:,:,:,i));
  fprintf(fid,',\n');
end

fprintf(fid,'\n');
fclose(fid);

function [index_c,index_d]=add_to_index(index_c,index_d,c1,c2,n)
% add boundary n to index array index at points c
for i=c1(1):c2(1)
  for j=c1(2):c2(2)
    for k=c1(3):c2(3)
      m=index_c(i,j,k); % How many points are there
      m = m + 1;
      index_c(i,j,k)=m; % Increase number of points
      index_d(i,j,k,m)=n; % Add boundary
    end
  end
end

% Print the elements of a structure out
function structfprintf(fid,s,i,ps)
range = [s.min s.max];
range(range==Inf) = 10^6;
range(range==-Inf) = -10^6;
fprintf(fid,[' &BOUNDARY_PARAM']);
fprintf(fid,['\n     BD_NAME    = "' s.name '"']);
fprintf(fid,['\n     BD_TYPE    = %i'],boundary_type(s.type));
fprintf(fid,['\n     BD_REGION  =']);fprintf(fid,[' ' ps],range);
fprintf(fid,['\n     BD_S       =']);fprintf(fid,[' ' ps],s.s);
fprintf(fid,['\n     BD_V       =']);fprintf(fid,[' ' ps],s.v);
fprintf(fid,'/\n\n');
