! Geometry module
! Contains functions for dealing with the geometry 
! including loading the geometry from a given file
! Calculates contacts and force posiibly using the contact forces module

! Different type of geometry at this time
  
module geometry
  use precision
  use vector
  use sysdep_io
  use program_units
  use particle
  use collision
  implicit none
  !private 
 
  ! Public functions
  public :: write_geometry, read_geometry, geom_get_extent
  public :: check_geom_bound, calc_pw, geom_info

  type geom_cylinder_t
     real(r_kind) length
     real(r_kind) radius
  end type geom_cylinder_t
  
  type geom_ski_slope_t
     real(r_kind) param_a
     real(r_kind) param_b
  end type geom_ski_slope_t

  ! Definitions of different types of geometry
  integer, parameter :: GEOM_CYLINDER = 1
  integer, parameter :: GEOM_CUBOID = 2 
  integer, parameter :: GEOM_SKI_SLOPE = 3
  
  
  type geom_t ! All in program units
     integer type      ! Type of geometry */
     integer n_wall    ! Number of different walls */
     type(rreg) bounds     ! bounding coordinates */
     type(rreg) cuboid   ! Just a region definition
     type(geom_cylinder_t) cylinder
     type(geom_ski_slope_t) ski
  end type geom_t

  ! All in SI units
  ! Geometry parameters for reading and writing files
  character(len=9) ::  geom_type = 'cylinder '  ! Geometry type
  real(r_kind):: geom_cylinder_length = 0.5             ! Length of cylinder 
  real(r_kind)    :: geom_cylinder_radius = 0.2         ! Radius of cylinder 
  type(rreg):: geom_cuboid_rg = rreg(rvec(.2,.2,.2),rvec(-.2,-.2,-.2))

  save:: geom_type,geom_cylinder_length,geom_cylinder_radius, geom_cuboid_rg
  ! The list of the variable names
  namelist /geometry_param/ geom_type,  &
       & geom_cylinder_length, geom_cylinder_radius, geom_cuboid_rg
  
  type(geom_t),save :: geom

contains
  subroutine geom_info(fp)
    integer, intent(in) :: fp
    write(fp,*) 'Geometry information'
  
    select case (geom%type)
    case(GEOM_CYLINDER)
       write(fp,*) 'Cylinder length radius : ', &
            geom%cylinder%length, geom%cylinder%radius
    case(GEOM_CUBOID)
       write(fp,*) 'Cuboid bounds :', geom%cuboid
    case(GEOM_SKI_SLOPE)
       write(fp,*) 'Ski slope not yet implimented'
    case default
       write(fp,*) 'Geometry type isn''t set'
    end select
    write(fp,*) 'Number of walls : ', geom%n_wall
    write(fp,*) 'Bounds : ', geom%bounds
  end subroutine geom_info
  
  function geom_get_extent()
    type(rreg) geom_get_extent 
    geom_get_extent = geom%bounds
  end function geom_get_extent

  ! Write geometry information to the file unit fp 
  subroutine write_geometry(fp)
    integer, intent(in) :: fp
    write(fp, nml=geometry_param)
  end subroutine write_geometry

  ! uses the data from datainout to set the geometry for a simulation
  subroutine read_geometry(fp)
    integer, intent(in) :: fp
    read(fp, nml=geometry_param)
    
    select case (geom_type)
    case('cylinder ')
       call set_cylinder
    case('cuboid   ')
       call set_cuboid
    case('ski_slope')
       call set_ski_slope
    case default
       call error_halt('Unknown geometry') 
    end select
    call assert(volume(geom%bounds)>0, 'Geometry has no volume')
  end subroutine read_geometry
  
  subroutine set_cylinder
    call info_msg('Geometry is cylinder radius')
     geom%type = GEOM_CYLINDER
     geom%n_wall = 3
     geom%cylinder%length = geom_cylinder_length
     geom%cylinder%radius = geom_cylinder_radius
     call conv_length_from_si(geom%cylinder%length)
     call conv_length_from_si(geom%cylinder%radius)
     geom%bounds%max = rvec(geom%cylinder%radius, geom%cylinder%radius,&
          & geom%cylinder%length/2)
     geom%bounds%min = -geom%bounds%max
   end subroutine set_cylinder

   subroutine set_cuboid
     geom%type = GEOM_CUBOID
     geom%n_wall = 6
     geom%cuboid = geom_cuboid_rg
     call conv_length_from_si(geom%cuboid%min)
     call conv_length_from_si(geom%cuboid%max)
     geom%bounds = geom%cuboid
   end subroutine set_cuboid
   
   subroutine set_ski_slope
     call error_halt('Set ski slope unimplimented')
   end subroutine set_ski_slope

   subroutine check_geom_bound(r)
     type(rvec), dimension(:), intent(in) :: r
     call assert(all(inside(geom%bounds,r)), 'Particles out of geometry')
   end subroutine check_geom_bound

   ! Calculates particle wall contacts and interactions */ 
   subroutine calc_pw(part) 
     type(part_list_t),intent(inout) :: part
     select case (geom%type)
     case (GEOM_CYLINDER)
        call calc_pw_cylinder(part)
     case (GEOM_CUBOID)
        call calc_pw_cuboid(part)
     case (GEOM_SKI_SLOPE)
        call calc_pw_ski(part)
     case default
       call error_halt('Unknown geometry') 
      end select
    end subroutine calc_pw

    subroutine calc_pw_cylinder(part)
      type(part_list_t),intent(inout) :: part
      type(rvec) :: r
      real(r_kind) :: rad, radsqr,len 
      integer i
      real(r_kind) ov ! overlap
      rad = geom%cylinder%radius-RADIUS
      radsqr = rad*rad
      len = geom%cylinder%length/2 - RADIUS
      do i = 1,size(part%p)
         r = part%p(i)%r
         ov = r%x**2 + r%y**2
         if (ov>radsqr) then  ! Side wall contact */
            ov = sqrt(ov)
            call spring_force (part%p(i), &
                 rvec(r%x/ov, r%y/ov,0), ov-rad)
         end if
         if (r%z > len) then ! End wall contact
            call spring_force (part%p(i), rvec(0,0,1), r%z-len)
         else if (-r%z > len) then  ! Contact */
            call spring_force (part%p(i), rvec(0,0,-1), -r%z-len)
         end if
      end do
    end subroutine calc_pw_cylinder
	
    subroutine calc_pw_cuboid(part)
      type(part_list_t),intent(inout) :: part
      type(rvec) :: r
      type(rreg) :: rg
      integer i
      rg = shrink(geom%cuboid,RADIUS)
      do i = 1,size(part%p)
         r = part%p(i)%r
         if (r%x < rg%min%x) then
            call spring_force (part%p(i), rvec(-1,0,0), rg%min%x-r%x)
         else if (r%x > rg%max%x) then
            call spring_force (part%p(i), rvec(1,0,0), r%x-rg%max%x)
         end if
         if (r%y < rg%min%y) then
            call spring_force (part%p(i), rvec(-1,0,0), rg%min%y-r%y)
         else if (r%y > rg%max%y) then
            call spring_force (part%p(i), rvec(1,0,0), r%y-rg%max%y)
         end if
         if (r%z < rg%min%z) then
            call spring_force (part%p(i), rvec(-1,0,0), rg%min%z-r%z)
         else if (r%z > rg%max%z) then
            call spring_force (part%p(i), rvec(1,0,0), r%z-rg%max%z)
         end if
      end do
    end subroutine calc_pw_cuboid
    subroutine calc_pw_ski(part)
      type(part_list_t),intent(inout) :: part
      part%p(1)%r=rvec(0,0,0)
      call error_halt('Ski module not yet written.')
    end subroutine calc_pw_ski
end module geometry   
