module wind
  use precision
  use sysdep_io
  use vector
  use program_units
  implicit none
  !private
  public :: init_wind, calc_wind_forces, read_wind_param, wind_info
  ! Parameters to be read in
  ! All in SI units
  real(r_kind),save :: air_density = 1.225       ! density Kg/m^3
  real(r_kind),save :: air_pressure = 0          ! pressure N/m^2
       ! real(r_kind), parameter :: PI =
       ! 3.141592653589793115997963468544
  real(r_kind),save :: air_viscosity = 0.0000145 ! Kinematic
       ! viscosity m^2/s
  real(r_kind),save :: air_temperature = 15      ! Degrees Centigrade 
  type(rvec),save :: air_velocity = rvec(0,0,0)
  namelist /air_param_nl/ air_density,air_pressure,air_viscosity&
       & ,air_temperature, air_velocity
  real(r_kind),save :: C1,C2,C3,C4       ! Coefficents for drag force
  real(r_kind),save :: terminal_velocity               ! Assuming real g
contains
  subroutine init_wind
  end subroutine init_wind
  ! Read the wind parameters
  ! And return an adjusted gravity that takes into account air buoyancy
  ! gravity is in SI units
  subroutine read_wind_param(fp)
    real(r_kind) p_density;
    type(rvec) :: gravity= rvec(0,0, -9.81)
    type(rvec) adj_gravity
    integer, intent(in) :: fp
    real(r_kind) :: Cd                 ! 1/s
    real(r_kind) :: Re_f            ! factor for calculating reynolds number
    real(r_kind) :: g
    real(r_kind) :: v(1)
    integer :: i
    read(fp, nml=air_param_nl) ! Read in the parameters
    call conv_from_si(air_density,T=0,L=-3,M=1)
    call conv_from_si(air_pressure,T=-2,L=-1,M=1)
    call conv_from_si(air_viscosity,T=-1,L=2,M=0)
    ! Convert to program units
    p_density = MASS/(.75*PI*RADIUS**3)     ! Particle density
    Re_f = 2*RADIUS/air_viscosity            ! Reynold number = Re_f * v
    Cd = 0.75*air_density/(2*RADIUS*p_density)
    C1 = Cd * 24./Re_f                    
    C2 = Cd * 6./sqrt(Re_f)               
    C3 = Cd * 1./sqrt(Re_f)               
    C4 = Cd * 0.4    
    adj_gravity = gravity*(1 -  air_density/p_density)
    g = length(adj_gravity)
    call conv_acceleration_from_si(g)
    v(1) = (sqrt(C1**2+4*g*C4)-C1)/(2*C4)
    do i = 1,5 ! This should give a very accurate result
       v = v-(g-v*drag_function(v))/D_drag_function(v) ! Use Newton Raphson
    end do
    terminal_velocity = v(1)
end subroutine read_wind_param
  function drag_function(speed) result(drag)  ! Make this easy to vectorise
    real(r_kind), intent(in) :: speed(:)
    real(r_kind) :: drag(size(speed))
    drag = C1 + (C2/(C3+sqrt(speed)) + C4)*speed
  end function drag_function
  function D_drag_function(speed) result(drag)  ! derivative of v times drag function
    real(r_kind), intent(in) :: speed(:)
    real(r_kind) :: drag(size(speed))
    drag = C1 + 2*(C2/(C3+sqrt(speed)) + C4)*speed - 0.5*C2*speed**1.5/(C3+sqrt(speed))**2
  end function D_drag_function
  function calc_wind_forces(r, v) result(f)
    type(rvec),intent(in) :: v(:),r(:)
    type(rvec) :: f(size(v))
    f = air_velocity-v                     ! Calculate difference
    f = f*drag_function(length(f))         ! Calculate force
  end function calc_wind_forces
  subroutine wind_info(fp)
      integer, intent(in) :: fp
      write(fp,*) 'Wind module information' 
      write(fp,*) 'air density', air_density
      write(fp,*) 'air pressure', air_pressure
      write(fp,*) 'air viscosity', air_viscosity
      write(fp,*) 'air velocity', air_velocity
      call conv_velocity_to_si(terminal_velocity)
      write(fp,*) 'SI terminal velocity', terminal_velocity
      call conv_velocity_from_si(terminal_velocity)
      write(fp,*) 'PU terminal velocity', terminal_velocity
      write(fp,*) 'C1', C1
      write(fp,*) 'C2', C2
      write(fp,*) 'C3', C3
      write(fp,*) 'C4', C4
  end subroutine wind_info
end module
