module boundary_f90_def
  use precision
  use vector
  use program_units

  ! Specification for general geometries
  integer, parameter :: NAME_L=20         ! Length of the geometry name
  integer, parameter :: MAX_S_PARAM_NO=5; ! maximum number of scalar parameters
  integer, parameter :: MAX_V_PARAM_NO=2; ! maximum number of vector parameters
  integer, parameter :: MAX_A_PARAM_NO=2; ! maximum number angular velocity parameters
  integer, parameter :: XP_PLANE   = 1;   ! x<0 is solid
  integer, parameter :: XM_PLANE   = 2;   ! x>0 is solid
  integer, parameter :: YP_PLANE   = 3;   ! y<0 is solid
  integer, parameter :: YM_PLANE   = 4;   ! y>0 is solid
  integer, parameter :: ZP_PLANE   = 5;   ! z<0 is solid
  integer, parameter :: ZM_PLANE   = 6;   ! z>0 is solid
  integer, parameter :: PLANE      = 7;
  integer, parameter :: P_CYLINDER = 8;
  integer, parameter :: M_CYLINDER = 9;
  integer, parameter :: R_CYLINDER = 10; ! A ring of cylinders
  integer, parameter :: R_PLANE    = 11; ! A plane of cylinders

  integer, parameter :: SPRING_BOUNDARY = 0;
  integer, parameter :: TRANSFORM_BOUNDARY = 1;

  interface conv_to_si
     module procedure conv_boundary_to_si_a
  end interface

  interface conv_from_si
     module procedure conv_boundary_from_si_a
  end interface

  ! Structure holding data for each boundary
  type boundary_t
     character(len=NAME_L) name       ! Boundary name
     integer type;                    ! Boundary curve TYPE
     integer sort;                    ! What SORT of boundary
     integer prop;                    ! Interaction PROPerties
     type(rreg_t) region;             ! region containing boundary not used
     real(r_kind) s(MAX_S_PARAM_NO)   ! scalar parameters
     type(rvec_t) v(MAX_V_PARAM_NO)   ! vectors parameters
     type(rang_t) w(MAX_A_PARAM_NO)   ! radius * angular velocity
     real(r_kind) theta               ! Angular position of cylinder
     real(r_kind) omega               ! Angular velocity of cylinder
     real(r_kind) aspace              ! Angular spacing of cylinders
     type(rvec_t) r_offset            ! position offset
     type(rvec_t) v_offset            ! velocity offset
  end type boundary_t

contains
  subroutine conv_boundary_to_si_a(b)
    type(boundary_t), intent(inout) :: b(:)
    integer i
    call conv_length_to_si(b%region%min) ! Convert units
    call conv_length_to_si(b%region%max) ! Convert units
    call conv_ang_vel_to_si(b%omega)
    do i=1,size(b)
       call conv_length_to_si(b(i)%s) ! Convert units
       call conv_velocity_to_si(b(i)%w) ! Convert units
    end do
  end subroutine conv_boundary_to_si_a

  subroutine conv_boundary_from_si_a(b)
    type(boundary_t), intent(inout) :: b(:)
    integer i
    call conv_length_from_si(b%region%min) ! Convert units
    call conv_length_from_si(b%region%max) ! Convert units
    call conv_ang_vel_from_si(b%omega)
    do i=1,size(b)
       call conv_length_from_si(b(i)%s) ! Convert units
       call conv_velocity_from_si(b(i)%w) ! Convert units
    end do
  end subroutine conv_boundary_from_si_a
end module boundary_f90_def

! R_CYLINDER has the following parameters
! The centre of the cylinder is at s(1)*v(1)
! The radius of the cylinder is s(2)
! v(2) is an orthonormal vector perpemdicular to the axis
! w(1) is the rotation of the main cylinder
! s(3) is the spacing of the auxillary cylinders
! s(4) is the radius of the auxillary cylinders
! w(2) is the extra rotation of the auxillary cylinders
