#include "part_props.h"
! Calculate a wall-particle spring interaction 
! Read in the data that this module needs and set the 
! appropriate variables
! This data is particle properties and collision properties
! and includes the adjusted gravity, taking into account air bouyancy
! pp means particlel-particle
! pw means particle-wall property
! Default is for frictionless elastic collisions


! Note that p%mp%r is angular_velocity*radius units m s^-1
! and that  p%am is angular_moment/radius       Kg m s^-2

module collision
  use sysdep_io
  use precision
  use vector
  use program_units
  use particle
  use col_props_f90_def
  use record
  use maths_const
  use maths
  use random
  implicit none
  !private SR2201BUG
  public :: write_part_param, read_part_param
  public :: spring_force, bcast_col_props
  public :: set_part_props

  interface spring_force
     module procedure pp_spring_force, pg_spring_force, pw_spring_force
  end interface

  ! Maximum number of different radius particles
  integer, parameter         :: MAX_P_NUM=20

  type(col_props_t), allocatable, private :: pp_col_props(:,:)
  type(col_props_t), allocatable, private :: pw_col_props(:,:)

  real(r_kind), allocatable, private :: part_radii(:)
  real(r_kind), allocatable, private :: part_masses(:)
  real(r_kind), allocatable, private :: part_mis(:)
  real(r_kind), allocatable, private :: part_dist(:)


  type(rvec_t), public :: gravity    ! Particle force due to gravity

contains
  ! Number of collision properties is hard coded in
  subroutine bcast_col_props
    use mpi
    integer s(4)
    call fmpi_bcast(gravity)
    if (fmpi_root) then
       s(1:2) = shape(pp_col_props)
       s(3:4) = shape(pw_col_props)
    end if
    call fmpi_bcast(s)
    if (.not.fmpi_root) then
       allocate(pp_col_props(s(1),s(2)), pw_col_props(s(3),s(4)))
    end if
    call fmpi_bcast(pp_col_props)
    call fmpi_bcast(pw_col_props)
  end subroutine bcast_col_props
  ! Calculate the forces between two particles and increase the forces
  ! for each particle
  subroutine pp_spring_force(p,q)
    type(particle_t), intent(inout) :: p,q
    type(rvec_t) tv ! Tangential velocity
    type(rvec_t) rv ! relative separation vector and half that 
    type(rvec_t) nrv! Unit separation vector 
    type(rvec_t) v  ! Relative velocity of centres
    type(rvec_t) f  ! Calculated force 
    type(rang_t) m  ! Calculated moment
    real(r_kind) r     ! separation distance 
    real(r_kind) ns    ! normal speed 
    real(r_kind) ov    ! 
    real(r_kind) cpe   ! potential energy stored in the contact 
    real(r_kind) n                        ! normal force
    type(col_props_t) cp
    cp = GET_PP_COLL_PROP(p%mp,q%mp)
    rv  = q%mp%r - p%mp%r  ! From p to q Separation vector
    r   = length(rv)             ! Separation distance     
    ov  = P_RADIUS(p%mp)+P_RADIUS(q%mp)-r                 ! overlap
    nrv = rv/r                   ! Unit separation vector 
    v   = p%mp%v - q%mp%v  ! Relative velocity of centres 
    ns  = nrv .dot. v            ! normal speed 
    tv  = v - ns * nrv + (nrv .x. (p%mp%w + q%mp%w)) ! tangential velocity  
    n = ov*cp%nk + ns*cp%neta
    f = n*(-nrv + friction_vec(tv,cp))
    if (rec_flag_energy()) then
       cpe = cp%nk * ov**2 / 4   ! Half the bond potential energy
       p%cpe = p%cpe + cpe ! Add on potential energy
       q%cpe = q%cpe + cpe ! Half for each particle
    end if
    p%f = p%f + f
    q%f = q%f - f
    m = nrv .x. f              ! Calculate moment 
    p%am = p%am - m 
    q%am = q%am - m
  end subroutine pp_spring_force
  subroutine pg_spring_force(p,q)
    type(particle_t), intent(inout) :: p
    type(min_part_t), intent(in) :: q
    type(rvec_t) tv ! Tangential velocity
    type(rvec_t) nrv! Unit separation vector 
    type(rvec_t) v  ! Relative velocity of centres
    type(rvec_t) f  ! Calculated force 
    real(r_kind)r     ! separation distance 
    real(r_kind)ns    ! normal speed 
    real(r_kind)ov    ! 
    real(r_kind)cpe   ! potential energy stored in the contact 
    real(r_kind) n                        ! normal force
    type(col_props_t) cp
    cp = GET_PP_COLL_PROP(p%mp,q)
    nrv  = q%r - p%mp%r  ! From p to q Separation vector
    r   = length(nrv)             ! Separation distance     
    ov  = P_RADIUS(p%mp)+P_RADIUS(q)-r                 ! overlap
    call assert(ov>=0,'Overlap negative in particle-particle collision')
    nrv = nrv/r                   ! Unit separation vector 
    v   = p%mp%v - q%v  ! Relative velocity of centres 
    ns  = nrv .dot. v            ! normal speed 
    tv  = v - ns * nrv + (nrv .x. (p%mp%w + q%w)) ! tangential velocity  
    n = ov*cp%nk + ns*cp%neta
    f = n*(-nrv + friction_vec(tv,cp))
    if (rec_flag_energy()) then
       cpe = cp%nk * ov**2 / 2   ! The bond potential energy
       p%cpe = p%cpe + cpe ! Add on potential energy
    end if
    p%f = p%f + f
    p%am = p%am - (nrv .x. f) ! Calculate moment 
  end subroutine pg_spring_force
  ! q is not a particle owned by this processor so there are no forces
  ! for it and all the energy needs to be stored here

  ! nrv is a unit normal vector away from the balls center
  ! ov is a positive overlap
  subroutine  pw_spring_force(p, nrv, ov, wv)
    type(particle_t), intent(inout) :: p
    type(rvec_t), intent(in)         :: nrv !normal unit vector away from centre
    real(r_kind), intent(in)        :: ov  ! amount of overlap
    type(rvec_t), intent(in), optional :: wv  ! wall velocity

    type(rvec_t) tv                        ! Tangential velocity
    real(r_kind)ns                        ! Normal speed 
    type(rvec_t) f                         ! Force on wall
    real(r_kind) n                        ! normal force
    type(rvec_t) v  ! Relative velocity of centres
    type(col_props_t) cp
    cp = GET_PW_COLL_PROP(p%mp,1)

    call assert(length(nrv)-1,'Not a unit vector')
    call assert(ov>=0,'overlap in pw_spring_force is negative')

    if (present(wv)) then
       v = p%mp%v - wv ! Relative velocity 
    else
       v = p%mp%v
    end if

    ns = nrv .dot. v                  ! Normal velocity
    tv = v - ns * nrv  + (nrv .x. p%mp%w) ! Tangential velocity

    n = ov*cp%nk + ns*cp%neta
    f = n*(-nrv + friction_vec(tv,cp))
    if (rec_flag_energy()) then
       p%cpe = p%cpe +  cp%nk * ov**2/2!Calculate the bond potential energy
    end if
    p%f  = p%f + f   ! Add on force 
    p%am = p%am - (nrv .x. f)   ! Add on moment scaled by particle radius 
  end subroutine pw_spring_force

  function friction_vec(tv,props) result(fv) ! Calculate the friction vector
    type(rvec_t), intent(in) :: tv
    type(col_props_t), intent(in) ::props
    type(rvec_t) :: fv
    real(r_kind) :: ts
    ts = length(tv) ! Tangential speed
    if (ts <= 0) then
       fv = null_rvec
    else if (ts<props%myu_vel) then  
       fv = (-props%myu/props%myu_vel) * tv
    else 
       fv = (-props%myu/ts) * tv 
    end if
  end function friction_vec


  ! The rest of the function are concerned with reading and writing the 
  ! data file and choosing the appropriate parameteres 

  ! read particle and collision properties
  ! choose program units
  ! returns gravity as second argument in program units

  subroutine read_part_param(fp,gout)
    integer, intent(in) :: fp
    type(rvec_t), intent(out) :: gout 

    !particle_param name list variables
    integer      :: p_num = 1
    real(r_kind) :: p_radius(MAX_P_NUM) = 1
    real(r_kind) :: p_mass(MAX_P_NUM) = 1
    real(r_kind) :: p_mi(MAX_P_NUM)=1!dimensionless moment of inertia I/(Mr^2)
    real(r_kind) :: p_dist(MAX_P_NUM)=1!unnormalized probability weights
    real(r_kind) :: p_max_vel   = 1
    real(r_kind) :: p_max_disp  = 0.01
    real(r_kind) :: p_time_step = 0 ! Should we force the time step
    real(r_kind) :: p_steps     = 10! Number of steps for fastest collision
    type(rvec_t) :: p_gravity    ! Particle force due to gravity
    type(col_props_t), allocatable  :: b_col_props(:)
    type(col_props_t), allocatable  :: p_col_props(:)
    type(col_props_t), allocatable  :: w_col_props(:)
    integer      :: num_p_props=1!Number of following particle collision props
    integer      :: num_w_props=1!Number of following wall collision props

    real(r_kind) min_r, max_r, max_m, omega
    integer i,j
    real(r_kind) :: pdt

    namelist /particle_param/ p_num, p_radius, p_mass, p_mi, p_gravity, &
         p_max_vel, p_max_disp, p_time_step, p_steps, num_p_props, &
         num_w_props, p_dist

    call info_msg('Reading particle and collision param.')    
    p_gravity = null_rvec
    read(fp, nml=particle_param) 

#ifdef ONE_TYPE
    if (p_num/=1) call error_halt('Program compiled for identical particles,&
         &but p_num in parameter file is not 1') 
#endif

    if (num_p_props>1) call error_halt('Only code for one set of particle &
         &collision properties is currently implemented')

    if (num_w_props>1) call error_halt('Only code for one set of wall &
         &collision properties is currently implemented')


    allocate(part_radii(p_num),part_masses(p_num),&
         part_mis(p_num),part_dist(p_num))
    part_radii = p_radius(1:p_num)
    part_masses = p_mass(1:p_num)
    part_mis = p_mi(1:p_num)*part_masses
    part_dist = p_dist(1:p_num)/sum(p_dist(1:p_num))

    max_r=maxval(part_radii)
    min_r=minval(part_radii)
    max_m=maxval(part_masses)

    if (max_r/min_r>RADIUS_RATIO) call error_halt('Radius ratio is too large')

    allocate(b_col_props(num_p_props+num_w_props)) ! Basic col_props
    do i = 1,size(b_col_props)
       b_col_props(i) = set_col_props(fp)
    end do
    ! This is the frequency for the system and inverse stiffness
    omega = p_max_vel/(p_max_disp*min_r)
    b_col_props%nk    = b_col_props%nk*omega**2
    b_col_props%neta = b_col_props%neta*omega

    ! No create a coll_prop for each particle type
    if (num_p_props==0) then
       allocate(p_col_props(0))
    elseif (num_p_props==p_num)  then
       allocate(p_col_props(p_num))
       p_col_props = b_col_props(1:num_p_props)
       p_col_props%nk   = p_col_props(1:num_p_props)%nk   * part_masses
       p_col_props%neta = p_col_props(1:num_p_props)%neta * part_masses
    elseif (num_p_props==1) then
       allocate(p_col_props(p_num))
       p_col_props = b_col_props(1)
       p_col_props%nk   = p_col_props(1)%nk   * part_masses
       p_col_props%neta = p_col_props(1)%neta * part_masses
    else
       call error_halt('collision.F90: num_p_props>1 and /=p_num')
    end if
    allocate(w_col_props(num_w_props))
    w_col_props  = b_col_props(num_p_props+1:num_p_props+num_w_props)

    num_p_props=size(p_col_props)

    ! Calculate particle particle collision info
    allocate(pp_col_props(p_num,p_num))
    do i=1,num_p_props
       do j=1,num_p_props
          pp_col_props(i,j) = add_col_props(p_col_props(i),p_col_props(j))
       end do
    end do

    ! Calculate particle wall collision info
    allocate(pw_col_props(p_num,num_w_props))
    do i=1,p_num
       do j=1,num_w_props
          pw_col_props(i,j)%nk      = p_mass(i) * w_col_props(j)%nk
          pw_col_props(i,j)%neta    = p_mass(i) * w_col_props(j)%neta
          pw_col_props(i,j)%myu     = w_col_props(j)%myu
          pw_col_props(i,j)%myu_vel = w_col_props(j)%myu_vel
       end do
    end do

    ! Choose the smallest time step
    pdt = min(real(0.01,r_kind),1/(omega*p_steps))

    if (p_time_step/=0) then 
       call warn_assert(p_time_step<pdt,'Forced time step is large')
       pdt = p_time_step
    end if

    ! Now set program units
    call set_units(pdt, max_r, max_mass)

    !Now scale everything to program units      
    call conv_from_si(pp_col_props)
    call conv_from_si(pw_col_props)
    call conv_acceleration_from_si(p_gravity)  ! Convert to program units
    call conv_length_from_si(part_radii)
    call conv_mass_from_si(part_masses)
    call conv_mass_from_si(part_mis)
    gout = p_gravity     ! Return gravity
    gravity = p_gravity  ! Global gravity 
  end subroutine read_part_param

  ! read the collision properties and calculate appropriate constants 
  function set_col_props(fp) result (cp)
    integer, intent(in)   :: fp 
    type(col_props_t) :: cp

    real(r_kind) :: stiffness =  1        ! extra stiffness 
    real(r_kind) :: restitution =  1      ! restitution coefficient 
    real(r_kind) :: friction = 0          ! friction coefficent 
    real(r_kind) :: friction_vel = 0      ! friction velocity
    character(len=2) :: type="  "         ! type of collision
  
    namelist /collision_param/ type, restitution, friction, friction_vel, &
         stiffness

    read(fp, nml=collision_param)
    cp%myu = friction
    cp%myu_vel = friction_vel
    !    cp%nk = (1 + (log(restitution) / PI)**2) * stiffness**2
    !    cp%neta = -2 * stiffness * log(restitution) / PI
    cp%nk = stiffness**2
    if (restitution<tiny(restitution)) then
       cp%neta = 2*stiffness
    else
       cp%neta = -2*stiffness*log(restitution)/sqrt(PI**2+log(restitution)**2)
    end if
  end function set_col_props

  subroutine write_part_param(fp)
    integer, intent(in) :: fp
    integer i,j
    do i=1,size(pp_col_props,1)
       do j=1,size(pp_col_props,2)
          write(fp,*)'pp_col_props(',i,',',j,')'
          call write_col_props(fp,pp_col_props(i,j))
       end do
    end do
    do i=1,size(pw_col_props,1)
       do j=1,size(pw_col_props,2)
          write(fp,*)'pw_col_props(',i,',',j,')'
          call write_col_props(fp,pw_col_props(i,j))
       end do
    end do
    call write_units(fp)
  end subroutine write_part_param

  subroutine write_col_props(fp,cpa)
    integer,intent(in)             :: fp
    type(col_props_t), intent(in) :: cpa
    type(col_props_t) :: cp
    real(r_kind)cpt_myu 
    real(r_kind)cpt_myu_vel 
    real(r_kind)cpt_nk
    real(r_kind)cpt_neta

    namelist /col_props_param/cpt_myu, cpt_myu_vel, cpt_nk, cpt_neta
    cp = cpa
    call conv_to_si(cp)
    cpt_myu      = cp%myu
    cpt_myu_vel  = cp%myu_vel
    cpt_nk       = cp%nk
    cpt_neta     = cp%neta
    write(fp,nml = col_props_param);
  end subroutine write_col_props  

  subroutine set_part_props(p)
    type(min_part_t), intent(inout) :: p(:)
    integer tps(size(p))
#ifdef PARTICLE_HAS_COL_PROPS
    call error_halt('Particle with collision properties is unimplimented')
#endif 
#ifdef PARTICLE_HAS_PHYS_PROPS
    tps = random_element(part_dist,size(tps))
    p%radius = part_radii(tps)
    p%imass  = 1/part_masses(tps)
    p%imi    = 1/part_mis(tps)
#endif
#ifdef PARTICLE_HAS_TYPE
    p%type   = tps
#endif 
  end subroutine set_part_props
  function add_col_props(a,b) result (r)
    type(col_props_t), intent(in) :: a,b
    type(col_props_t) :: r
    r%nk = harmonic_mean(a%nk,b%nk)
    r%neta = harmonic_mean(a%neta,b%neta)
    r%myu = sqrt(a%myu*b%myu)
    r%myu_vel = sqrt(a%myu_vel*b%myu_vel)
  end function add_col_props
end module collision
