#include "dimension.h"
#include "part_props.h"
! Takes it's input from std_input so we can pipe in files
module def_particles
  use particle
  type(part_list_t) part ! Main list of particles 
  type(min_part_list_t) ghost_part ! List of ghost particles
end module def_particles

program dem
  use precision
  use sysdep_io
  use vector
  use geometry
  use pd
  use grid
  use program_units
  USE def_particles  ! Stop auto dependency generation
  use collision
  use initial_state
  use wind
  use locate
  use record
  use termination
  use comms
  use pro_state_f90_def
  use stats
  use matrix
  USE f90_gc
  implicit none

  type(pro_state_t)  :: pro_state  ! State of this processor
  type(pro_state_t)  :: all_state  ! Collected state of all processors
  type(pro_state_t), allocatable :: all_pro_state(:) ! States of all the processors

  type(locate_t) :: pod ! processor owner distribution
  type(locate_t) :: pad ! processor array distribution
  type(min_part_list_t) init_part ! Initial list of particles

  integer, parameter :: FLENGTH=100    ! Longest file name 
  character(len=FLENGTH) :: pname

  logical, parameter :: debug_pro_dist = .false. ! Debug processor distribution
  logical, parameter :: debug_flow = .true. ! Debug program flow

  ! Statistics for how many particle this processor owns 
  type(stats_i_t) :: p_num_stats
  type(stats_i_t) :: g_num_stats  ! Ghost particle stats
  type(stats_i_t) :: x_num_stats  ! Exchange particle stats

  character(len=FILE_NAME_LEN) geom_file_name  ! File name for geometry file
  logical  :: dynamic_grid      = .true. 
  logical  :: grid_geom_bounds  = .false.

  namelist /program_param/ geom_file_name, dynamic_grid, grid_geom_bounds

  SILENT_GC =.TRUE. ! Debug garbage collection

  call stats_clear(p_num_stats)
  call stats_clear(g_num_stats)
  call stats_clear(x_num_stats)

  call comms_initialize                 ! Start up MPI

  call assign_lun(std_log) ! Get a unit number
  write(pname,'(i1)')comms_rank
  open(std_log, file='logs/p'//trim(pname)//'.log', status='replace', action='write')

  !std_out = std_log

  call fix_random_numbers !unrandomise random numbers

  write(std_log,*)'Fortran 90 parallel dem version 0.1'
  write(std_log,*)'Running on processor',fmpi_rank,'of',fmpi_size

  if (fmpi_root) then        
     call read_param(init_part)    
     call write_geometry('dem.geom')
     call write_param('dem.calc')
     call save_part_props(init_part%p)
     allocate(all_pro_state(fmpi_size))
  else  
     allocate(all_pro_state(0))
     call alloc(init_part,0)   
  end if

  ! make processor distribution array with default shape
  ! These next to statements just define the distribution
  ! The actual assignment is of regions is calculate every program step
  pod = make_processor_locate_t(comms_shape-scalari_to_ivec(1)) 
  ! Make locate type for ghost particle distribution
  pad = make_adjacent_locate_t(pod)
  ! Now we have to make sure all the necessary data is distributed 
  ! onto each nod
  call broadcast_all()
  !call write(std_log,pod,'Initial pod')
  !call write(std_log,pad,'Initial pad')
  ! Read or calculate the initial particle positions
  ! and distribute evenly to all processors

  !call particles_info(init_part, std_log)
  call initialize_particles(part,init_part)

  !call geom_info(std_log)
  !call particle_position_info(part, std_log)

  call simulate
  call end_grid(std_out)   

  call stats_print(std_log, g_num_stats,'Ghost particle sent')
  call stats_print(std_log, x_num_stats,'Particles transfered')
  call stats_print(std_log, p_num_stats,'Particles owned')

  close(std_log)
  call comms_finalize
contains
  subroutine initialize_particles(p,ip)
    type(part_list_t),intent(out) :: p ! particle list to return
    type(min_part_list_t),intent(inout) :: ip ! Initial list of particles
    type(min_part_list_t) recv_part ! Receive list of particles
    integer :: ppp                  ! particles per processor
    integer :: tp                   ! total particle
    integer :: n
    integer :: count(fmpi_size)    ! Number of particles per processor
    if (fmpi_root) then ! Only run this on one node
       tp = size(ip%p)
       ppp = tp/fmpi_size      ! Share them out evenly
       n = fmpi_size-(tp - fmpi_size*ppp) ! particle left over
       count(1:n) = ppp      ! particles 1 totp-rp get ppp 
       count((n+1):fmpi_size) = ppp+1 ! particles tp-rp+1 to tp  get ppp
       write(std_log,*)'Root distributing particles',count
    end if
    call fmpi_scatter(count,ppp) 
    call fmpi_bcast(tp) 
    call alloc(recv_part,ppp,ppp)
    call fmpi_scatterv(ip%p,count,displ_sum(count),recv_part%p)
    call dealloc(ip)
    write(std_log,*)'Receiving',ppp,'particles'
    write(std_log,*)'Allocating space for up to',tp
    call alloc(p,ppp,tp)  ! Allocate lots of extra space
    p = recv_part ! Copy into particle structure
    call dealloc(recv_part)
    p%p%nc = 0                    ! Initially no contacts
    p%p%cpe = 0                   ! Zero potential energy
    p%p%f = null_rvec          ! No force
    p%p%am = null_rang         ! No angular momentum
    p%p%grid = -1                 ! No grid square
    call fmpi_barrier
    write(std_log,*)'Particles initialised'
  end subroutine initialize_particles

  subroutine simulate
    real(r_kind)   :: t        ! Program time SI units 
    integer        :: step     ! Step number, program time program units

    ! Do we fix the grid to cover the geometry or allocate it dynamicaly
    if (grid_geom_bounds) then
       call init_grid(grow(geom_get_extent(),4*2*max_radius))
    else
       call init_grid()
    end if

    if (debug_flow) then
       call fmpi_barrier
       write(std_log,*)'Grid initialized'
    end if
    step = 0
    ! The first step is step zero and consists of adjusting the velocities back in time
    ! to v(-1/2dt) and w(-1/2dt)
    do     
       if (one_step(step)) exit               ! Execute one time step
       step = step+1 
    end do
    call end_record         ! Finish writing any records
    t = DT*step
    call conv_time_to_si(t)              ! Convert to SI units
    write(std_log,*)'Ending at time',t,'step',step
    call save_state(part) ! And dump the whole state
  end subroutine simulate
  ! do we make termination cooperative or non cooperative ?
  ! simplest for now is to leave it non cooperative
  ! and just decided on the collected data in all_pro_state
  ! Execute one time step on one processor
  ! At the end of this function position is given at time t
  ! and the velocities are given at time t-1/2dt
  function one_step(step) result(exit_flag)
    integer,intent(in)   :: step     ! Step number, program time program units
    real(r_kind)   :: t        ! Program time SI units 
    logical :: exit_flag       ! And of local exit flags on all nodes
    logical :: local_exit_flag ! Does this node think we should exit ? 
    if (fmpi_root) then 
       t = DT*step
       call conv_time_to_si(t)              ! Convert to SI units
    end if
    local_exit_flag = termination_criterion(part%p,step,t)
    call fmpi_allreduce(local_exit_flag,exit_flag,FMPI_LAND)
    ! If we're terminating force a record
    call start_record(step,t,exit_flag)
    ! Decide what stuff we want to record this step
    ! If the processor shape changes need to do more
    if (dynamic_grid) pro_state = calc_pro_state(part%p)
!!$    if (fmpi_size>1) then
!!$       call fmpi_gather(pro_state, all_pro_state) ! collect them together
!!$       if (fmpi_root) then ! Root processor does allocation
!!$          call calc_pod(all_state, pod, all_pro_state)
!!$       end if
!!$       call fmpi_bcast(pod%allc)  ! Broadcast the new cut positions
!!$       pad%allc=make_adjacent_cuts(pod%allc) ! Update the adjacent cuts
!!$       !call write(std_log,pad,'updated pad')
!!$       ! call restrict_locate_t(pod, pro_state%region)
!!$       ! call restrict_locate_t(pad, pro_state%region)
!!$       ! This is the spot for particle exchange
!!$       call particle_exchange(part,ghost_part,pad)
!!$       ! Make sure the grid region includes the ghost particles and the 
!!$       ! new particles
!!$       pro_state%region = union(region(part%p%mp%r),region(ghost_part%p%r))
!!$    end if
!!$    ! Move the grid by the average velocity
    if (dynamic_grid) call offset_grid(pro_state%v_mean)
    !Are the particles in the region ?
    if (.not.grid_geom_bounds) then
       if (.not.grid_contains(pro_state%region)) then
          call init_grid(grow(pro_state%region,4*2*max_radius)) ! Generate a new grid
       end if
    end if
    call init_forces(part)        ! Set forces to gravity
    call calc_wind_forces(part%p%f, P_RADIUS(part%p%mp), part%p%mp%v) ! Add on wind force
    part%p%grid=coord_to_grid(part%p%mp%r)
    call add_to_grid(part%p%grid)
    call calc_pp_contacts(part)   ! Calculate particle particle contacts
!!$    if (fmpi_size>1) then
!!$       call calc_pg_contacts(part,ghost_part) ! Particle ghost contacts and forces
!!$    end if
    call zero_grid(part%p%grid)   ! Clear the grid squares now
    call calc_pp_forces(part)     ! Calculate particle particle forces
    call calc_pw(part)            ! Calculate particle wall interactions

    if (step/=0) then               !
       call calc_displacements(part) ! Integrate forwards 
    else
       call offset_velocities(part) ! step velocities back dt/2 
    end if
    call displace_boundaries() ! Update the boundaries
    call make_record(part, step, t)! Record necessary data
    !call check_particles(part)    ! Check the particles are in sensible places
    !call assert(isfinite(part%p),'Something in the particles is infinite')
  end function one_step
  ! Put in abarrier and print a string on the root node
  subroutine df(s)
    character(len=*), intent(in) :: s
    if (debug_flow) then
       call fmpi_barrier
       if (fmpi_root) write(std_log,*)s
    end if
  end subroutine df
  ! Uses 
  !      pod : processor owner locate type
  !      pad : processor user locate type
  !     part : list of particles on this processor
  ! Builds list of which particles have to go where
  ! and exchanges them
  ! There are a lot of different ways of trying to do this efficently
  ! Define a lot of local variables and hope the compiler 
  !  can make sense of them
  ! ghost_part is a global allocatable particle array
  ! no signifies are variable to do with New Owner particles
  ! us signifies a variable to do with USe particles
  ! How about counting number of particle exchanges ?
  subroutine particle_exchange(p,up,pad)
    type(part_list_t), intent(inout) :: p ! Current particles 
    type(min_part_list_t) :: up    ! particle we need to use
    type(locate_t), intent(in) :: pad        ! processor distribution

    integer :: no_pn(size(p%p))   ! new owner particle number 
    integer :: no_dest(size(p%p)) ! new owner destination node 
    integer :: no_send_total      ! Total own particles sending
    integer :: no_recv_total       ! Total own particles receiving
    integer :: no_send_count(0:fmpi_size-1)
    integer :: no_send_displ(0:fmpi_size-1)
    integer :: no_recv_count(0:fmpi_size-1)
    integer :: no_recv_displ(0:fmpi_size-1)
    type(min_part_t), allocatable :: own_part(:) ! Own particles we receive


    !! user arrays could be upto 8 times the number of particles !
    integer :: gh_pn(8*size(p%p))     ! user particle number
    integer :: gh_dest(8*size(p%p))   ! user particle number
    integer :: gh_send_total        ! Total use particles sending
    integer :: gh_recv_total        ! Total use particles receiving
    integer :: gh_send_count(0:fmpi_size-1)
    integer :: gh_send_displ(0:fmpi_size-1)
    integer :: gh_recv_count(0:fmpi_size-1)
    integer :: gh_recv_displ(0:fmpi_size-1)
    integer, pointer :: d(:)
    integer :: i,j,n
    type(min_part_t), allocatable :: no_part(:), gh_part(:)

    logical, parameter :: debug_p_exch = .true.

    no_send_count = 0 ! New owner counts
    gh_send_count = 0 ! Ghost particle counts
    no_send_total = 0
    gh_send_total = 0

    do i=1,size(part%p)
       call get_data(d,pad,p%p(i)%mp%r) !d points to the data at a point
       ! d(1) contains owning processor
       ! d(2:) contains all the processors that need this particle
       ! d(2:) must (and should) be distinct
       if (d(1) /= fmpi_rank)  then ! Is this our region?
          no_send_count(d(1)) = no_send_count(d(1))+1
          no_send_total = no_send_total+1 
          no_pn(no_send_total) = i
          no_dest(no_send_total) = d(1)
       end if
       n = size(d)-1 ! How many 
       if (n<1) cycle ! Are there any ghost regions 
       gh_send_count(d(2:))=gh_send_count(d(2:))+1 
       gh_pn(gh_send_total+1:gh_send_total+n)=i    
       gh_dest(gh_send_total+1:gh_send_total+n)=d(2:) 
       gh_send_total = gh_send_total + n
    end do
    call assert(sum(gh_send_count)==gh_send_total, &
         'sum(gh_send_count) /= gh_send_total')
    call assert(sum(no_send_count)==no_send_total, &
         'sum(no_send_count) /= no_send_total')
    ! Allocate space for the temporary lists
    ! Possibly do something cleverer with particles we no longer own
    ! but still use
    allocate(no_part(no_send_total))
    allocate(gh_part(gh_send_total))
    no_send_displ = displ_sum(no_send_count)
    no_recv_count = no_send_displ
    gh_send_displ = displ_sum(gh_send_count)
    gh_recv_count = gh_send_displ
    do i = 1,no_send_total    ! Add all the particles in the right places
       j = no_dest(i)
       no_recv_count(j)=no_recv_count(j)+1
       no_part(no_recv_count(j)) = p%p(no_pn(i))
    end do
    !SR2201BUG    call assert(no_recv_count-no_send_count==no_send_displ, &
    !         'New owner particle counts incorrect')
    do i = 1,gh_send_total! Add all the particles in the right places
       j = gh_dest(i)
       gh_recv_count(j)=gh_recv_count(j)+1
       gh_part(gh_recv_count(j)) = p%p(gh_pn(i))
    end do
    !SR2201BUG    call assert(gh_recv_count-gh_send_count==gh_send_displ, &
    !         'Use particle counts incorrect')

    call fmpi_alltoall(no_send_count,no_recv_count)
    call remove(part,no_pn(1:no_send_total))! Remove the particle we no longer own
    no_recv_displ = displ_sum(no_recv_count)
    no_recv_total = no_recv_displ(fmpi_size-1)+no_recv_count(fmpi_size-1)
    ! More efficent is to copy this straight in from MPI
    allocate(own_part(no_recv_total))  
    call fmpi_alltoallv(no_part,no_send_count,no_send_displ, &
         own_part,no_recv_count,no_recv_displ)
    call push(p,own_part) ! Copy them onto the list
    deallocate(own_part)

    call fmpi_alltoall(gh_send_count,gh_recv_count)
    gh_recv_displ = displ_sum(gh_recv_count)
    gh_recv_total = gh_recv_displ(fmpi_size-1)+gh_recv_count(fmpi_size-1)
    call alloc(up,gh_recv_total) ! Allocate use particles
    call fmpi_alltoallv(gh_part,gh_send_count,gh_send_displ, &
         up%p,gh_recv_count,gh_recv_displ)
    call stats_add(g_num_stats,gh_send_total)
    call stats_add(x_num_stats,no_send_total)
    call stats_add(p_num_stats,size(p%p))
  end subroutine particle_exchange

  ! calculate processor owner distribution
  ! Given information about different processors calculate a distribution
  ! of space between them 
  ! Don't do much at the moment
  ! divide up the space evenly between all processors
  subroutine calc_pod(as, lt, s)
    type(pro_state_t), intent(out):: as
    type(locate_t), intent(inout) :: lt
    type(pro_state_t), intent(in) :: s(:)
    as = reduce(s)
    call lin_cut(lt%xc,as%region%min%x,as%region%max%x)
    call lin_cut(lt%yc,as%region%min%y,as%region%max%y)
    CMD3(call lin_cut(lt%zc,as%region%min%z,as%region%max%z))
  end subroutine calc_pod
  subroutine lin_cut(x,minval,maxval)
    real(r_kind), intent(inout) :: x(:)
    real(r_kind), intent(inout) :: minval, maxval
    integer i
    real(r_kind) t
    if (size(x)==0) return
    t = real(maxval-minval,r_kind)/real(size(x)+1,r_kind)
    if (t < 2*max_radius) t = 2*max_radius
    x = (/(minval+i*t,i=1,size(x))/)
  end subroutine lin_cut
  ! Calculate information about the particles on this node
  ! for load balancing
  ! Also processor load and memory usage
  function calc_pro_state(p) result(r)
    type(particle_t), intent(in) :: p(:)
    type(pro_state_t) :: r
    r%num_p = size(p)
    r%load = 1.0 ! Set all the loads equal to one for now
    r%g_mem = 0 ! Ignore this one as well at the moment
    r%region = region(p%mp%r) ! Region the particles are in
    if (r%num_p/=0) then
       r%r_mean = sum(p%mp%r)/ r%num_p 
       r%v_mean = sum(p%mp%v)/ r%num_p 
       r%r_sqr = inertia_matrix(p%mp%r) / r%num_p
    else
       r%r_mean = null_rvec
       r%v_mean = null_rvec
       r%r_sqr = null_rmat
    end if
  end function calc_pro_state
  subroutine write_param(fname)
    character(len=*), intent(in) :: fname
    integer fp
    call info_msg('Writing parameters to ' // trim(fname))
    call assign_lun(fp) ! Get a unit number
    open(fp,file=fname, delim='quote')
    call write_part_param(fp)
    close(fp)
  end subroutine write_param
  subroutine read_param(ip)
    type(min_part_list_t) :: ip
    integer fp  
    type(rvec_t) :: gravity

    call info_msg('Reading parameters from standard input')
    call assign_lun(fp); ! Get a unit number
    fp = std_in
    read(fp, nml=program_param)

    call read_record_param(fp)       ! what to record and how often
    call read_part_param(fp,gravity) ! Particle parameters
    call read_wind_param(fp,gravity) ! read wind parameters
    call read_termination_param(fp)  ! when to finish the simulation
    call read_geometry(trim(geom_file_name))
    !    call info_initial_state(fp, std_log)  ! Display info about initial state
    call read_initial_state(fp, ip)  ! Read/calculate initial particles
    close(fp)
  end subroutine read_param
  subroutine broadcast_all
    real(r_kind) u(3) ! Array for units
    call bcast_wind ! Broadcast wind parameters
    call bcast_col_props      ! Broadcast the collision properties
    call bcast_geometry        ! geometry information
    if (fmpi_is_root()) u = get_units()
    call fmpi_bcast(u)
    if (.not.fmpi_is_root()) call set_units(u(1),u(2),u(3)) 
  end subroutine broadcast_all
  subroutine fix_random_numbers
    integer n
    integer, allocatable :: seed(:)
    call random_seed(n)
    allocate(seed(n))
    call random_seed(get=seed)   ! old seed
    seed = 97284202              ! Choose all the same
    call random_seed(put=seed)   ! new seed
    deallocate(seed)
  end subroutine fix_random_numbers
  ! create a locate type that assigns a process to each region
  function make_processor_locate_t(nc) result(lt)
    type(ivec_t), intent(in) :: nc
    type(ivec_t) :: nnc
    type(locate_t) :: lt
    integer i,v
    call allocate(lt,nc,1) ! Create the structure
    lt%max_n = 1       ! every region has one processor
    nnc = nc + scalari_to_ivec(1)
    v = volume(nnc)
    lt%n = 1
    lt%d(ARG(:,:,:),1) = reshape((/(i,i=0,v-1)/),vec_to_array(nnc))
    if (debug_pro_dist) then
       call write(std_log,lt,'make_processor_locate()')
    end if
  end function make_processor_locate_t
  ! returns a real array three times the size with the elements
  ! r(3i-2)=a(i)-2*max_radius
  ! r(3i-1)=a(i)
  ! r(3i  )=a(i)+2*max_radius
  ! where i = 1,size(a)
  function make_adjacent_cuts(a) result(r)
    real(r_kind), intent(in) :: a(:)
    real(r_kind) :: r(3*size(a))
    integer i
    do i = 1,size(a)
       r(3*i-2) = a(i)-2*max_radius
       r(3*i-1) = a(i)
       r(3*i)   = a(i)+2*max_radius
    end do
  end function make_adjacent_cuts
  ! create a locate type that says which processors need to see a 
  ! particlular particle.
  ! takes the processor locate type from above as the template
  ! The owner and user for a region should always be different
  function make_adjacent_locate_t(in_lt) result(out_lt)
    type(locate_t), intent(in) :: in_lt
    type(locate_t) :: out_lt
    type(ivec_t) :: nc 
    integer max_n,ARG(i,j,k),ARG(id,jd,kd),ARG(im,jm,km),ARG(ii,jj,kk),ARG(is,js,ks)
    nc = 3*in_lt%nc
    ! Calculate the maximum number of involved processors
    max_n=1
    if (nc%x>0) max_n = 2*max_n
    if (nc%y>0) max_n = 2*max_n
    CMD3(if (nc%z>0) max_n = 2*max_n)
    call allocate(out_lt,nc,max_n) ! Create the structure
    out_lt%xc=make_adjacent_cuts(in_lt%xc)
    out_lt%yc=make_adjacent_cuts(in_lt%yc)
    CMD3(out_lt%zc=make_adjacent_cuts(in_lt%zc))

    out_lt%n=0
    ! id,jd,kd are coordinates on underlying grid specifying owning processor
    ! im,jm,km are -1,0,+1 specifying offset to surrounding processor
    do i=0,nc%x
       id = (i+1)/3;im = i - 3*id; 
       do j=0,nc%y
          jd = (j+1)/3;jm = j - 3*jd; 
#ifdef THREEDIM
          do k=0,nc%z
             kd = (k+1)/3;km = k - 3*kd;
             ks = 1 + km*(1-km)  ! map -1 => -1 , (0,1) => 1
#endif
             ! Maybe using a lookup array or if statements is faster
             ! this seems neatest
             is = 1 + im*(1-im)  ! map -1 => -1 , (0,1) => 1
             js = 1 + jm*(1-jm)  ! map -1 => -1 , (0,1) => 1
             do ii = 0,im,is
                do jj = 0,jm,js
#ifdef THREEDIM
                   do kk = 0,km,ks
#endif
                      call push(in_lt%d(ARG(id+ii,jd+jj,kd+kk),1),out_lt,ARG(i,j,k))
                      CMD3(end do)
                   end do
                end do
                CMD3(end do)
             end do
          end do
          if (debug_pro_dist) then
             call write(std_log,out_lt,'make_adjacent_locate()')
          end if
        end function make_adjacent_locate_t
      end program dem

! Subroutine that is called if an error occurs should save the state of all
! variables to a file.
! we don't know the time or step number
! doesn't seem to be a way to call back to the main program
subroutine error_save() 
  use precision
  use sysdep_io
  use vector
  use geometry
  use pd
  use grid
  use program_units
  USE def_particles ! Stop auto dependency generation
  use collision
  use initial_state
  use wind
  use locate
  use record
  use termination
  implicit none
  integer fp
  integer i
  call assign_lun(fp) ! Get a unit number
  open(fp,file='dem.error')
  if (associated(part%p)) then
     write(std_err,*)'Saving particles'
     do  i=1,size(part%p)
        call print_particle(fp,part%p(i))
     end do
     ! ordinary program end
     write(std_err,*)'Ending records'
     call end_record ! Finish ordinary records
     ! save_state probably messes up units?? fix this ??
     !call save_state(part, 'dem.dump', -1, -1.0) 
     call error_record(part,fp)! We don't know the time or step here
  else
     write(fp,*)'Error occured before particle list was created.'
     write(std_err,*)'Error occured before particle list was created.'
  end if
  close(fp)
end subroutine error_save

