#include "dimension.h"
#include "part_props.h"
! Geometry module
! Contains functions for dealing with the geometry 
! including loading the geometry from a given file
! Calculates contacts and force posiibly using the contact forces modulewr
 
  
module geometry
  use precision
  use vector
  use sysdep_io
  use program_units
  use particle
  use collision
  use locate
  use mpi
  use boundary_f90_def
  implicit none
  !private SR2201BUG 

  ! Public functions
  public :: write_geometry   ! Write a geometry description file 
  public :: read_geometry    ! Read a geometry description file
  public :: geom_get_extent  ! Return geometry bounds
  public :: check_geom_bound ! Check particles and print error description
  public :: calc_pw          ! Calculate particle=wall forces
  public :: geom_info        ! print information about the Geometry
  public :: bcast_geometry   ! Broadcast geometry info to all processors
  public :: displace_boundaries ! Integrate boundaries forward in time

  ! Different walls can be of different types and walls of the same type
  !              Can also have different properties friction etc
  ! What sorts of boundaries are interesting ?
  ! Ordinary wall type boundaries possibly moving spring or other interactions
  !                 Interaction possibilies include roughness
  !                 Can be regarded as variation of normal and overlap distance
  ! Mirror boundary particle interacts with a mirror particle
  !                 the same as a frictionless wall ?
  ! Periodic boundary Particles are translated, 
  !                  rotated including spin and velocity


  ! Variables
  type(boundary_t), allocatable :: boundaries(:)  !boundary data 
  type(locate_t)         :: geom_loc  ! Structure for indexing geometry

  ! Geometry parameters 
  integer               :: geom_num    = -1  ! Number of boundaries
  type(rreg_t)          :: geom_bounds       ! overall bounds
  character(len=NAME_L) :: geom_type         ! Geometry type
  character(len=NAME_L) :: geom_name         ! Geometry name
  real(r_kind)          :: geom_radius = -10 ! Geometry design radius
  integer               :: geom_dim    = DIM ! How many dimensions designed for
  character(len=NAME_L) :: matlab_name ="" 
  ! What matlab command or file to use to plot the geometry  

  namelist /geometry_param/ geom_type, geom_name, geom_num, geom_radius, &
       geom_bounds, geom_dim

  ! Namelist parameters for each individual boundary
  real(r_kind)          bd_s(MAX_S_PARAM_NO) ! scalar parameters
  type(rvec_t)          bd_v(MAX_V_PARAM_NO) ! vectors parameters
  real(r_kind)          bd_w(MAX_A_PARAM_NO) ! Angular velocity
  real(r_kind)          bd_theta             ! Angular position
  type(rreg_t)          bd_region            ! region
  integer               bd_type              ! type
  character(len=NAME_L) bd_name              ! Boundary name

  namelist /boundary_param/ bd_name, bd_type, bd_region, bd_s, bd_v, &
       bd_w, bd_theta 

  logical, private, parameter :: debug_geom = .false. !debugging messages

contains

  subroutine bcast_geometry
    call comms_bcast(geom_loc)
    call fmpi_bcast(geom_num)
    call fmpi_bcast(geom_bounds)
    if (.not.fmpi_root) allocate(boundaries(geom_num))
    call fmpi_bcast(boundaries)
  end subroutine bcast_geometry

  function geom_get_extent()
    type(rreg_t) geom_get_extent 
    geom_get_extent = geom_bounds
  end function geom_get_extent

  ! Write geometry information to the file fname
  subroutine write_geometry(fname)
    character(len=*), intent(in) :: fname
    integer fp
    integer i
    call assign_lun(fp); ! Get a unit number

    call conv_to_si(boundaries)
    call conv_to_si(geom_loc) ! Convert units
    call conv_length_to_si(geom_bounds%min)
    call conv_length_to_si(geom_bounds%max)

    open(fp,file=fname, delim='quote') 
    write(fp, nml=geometry_param)
    do i=1,geom_num
       call write_boundary(fp,boundaries(i)) 
    end do
    call write(fp, geom_loc)
    close(fp)

    call conv_from_si(boundaries)
    call conv_from_si(geom_loc) ! Convert units
    call conv_length_from_si(geom_bounds%min)
    call conv_length_from_si(geom_bounds%max)

  end subroutine write_geometry
  ! Integrate boundaires forward in time
  subroutine displace_boundaries()
    boundaries%theta = mod(boundaries%theta+DT*boundaries%omega,2*pi) 
  end subroutine displace_boundaries

  ! uses the data from data inout to set the geometry for a simulation
  subroutine read_geometry(fname)  
    character(len=*), intent(in) :: fname
    integer fp, i
    call info_msg('Reading geometry from ' // trim(fname))
    call assign_lun(fp); ! Get a unit number
    open(fp, file=fname, delim='quote', action='read', status='old')
    geom_bounds = null_rreg
    read(fp, nml=geometry_param)
    if(debug_geom) write(std_out,*)'geom_num',geom_num
    
    if (geom_dim /= DIM) &
         call error_halt('Geometry file is not for DIM dimensions')  
    if (geom_num<0 .or. geom_num >10000) &
         call error_halt('geometry.f90 geom_num is out of range')
    allocate(boundaries(geom_num))
    do i=1,geom_num
       call read_boundary(fp,boundaries(i)) 
    end do
    call info_msg('Read boundaries',debug_geom)
    call read(fp,geom_loc)
    call info_msg('Read locator',debug_geom)
    close(fp)
!    if (geom_radius /= -1) then ! -1 is special value any radius is ok
!       call conv_length_from_si(geom_radius)
!       if (geom_radius/=RADIUS) call info_msg('Geometry radius is incorrect',&
!            data=geom_radius)
!    end if
    call conv_from_si(boundaries)
    call conv_from_si(geom_loc) ! Convert units

    call conv_length_from_si(geom_bounds%min)
    call conv_length_from_si(geom_bounds%max)

    call check_geometry
  end subroutine read_geometry       
  subroutine read_boundary(fp, b)
    integer, intent(in) :: fp
    type(boundary_t), intent(out) :: b
    bd_name = 'none'
    bd_type = -1
    bd_s = 0
    bd_v = null_rvec
    bd_w = 0 
    bd_theta = 0
    read(fp,nml=boundary_param)
    b%name = bd_name
    b%type = bd_type
    b%region = bd_region  
    b%s = bd_s
    b%v = bd_v
    b%theta = bd_theta
    b%omega = bd_w(1)
    b%w(1) = bd_w(1)*b%s(2)*(b%v(1).x.b%v(2))
    if (b%type == R_CYLINDER) then
       b%w(2) = ((bd_w(1)+bd_w(2))*b%s(4))*(b%v(1).x.b%v(2))
       b%aspace = 2*pi/ceiling(pi/asin(b%s(3)/b%s(2)));!angle between cylinders
       b%s(3) = b%s(2) * sin(b%aspace/2);
       b%s(5) = b%s(2) * cos(b%aspace/2);    ! Calculate distance to mid point
    end if
  end subroutine read_boundary
  subroutine write_boundary(fp, b)
    integer, intent(in) :: fp
    type(boundary_t), intent(in) :: b
    bd_name = b%name
    bd_type = b%type
    bd_region = b%region    
    bd_s = b%s 
    bd_v = b%v
    bd_w = length(b%w)
    write(fp,nml=boundary_param)
  end subroutine write_boundary

  subroutine check_geom_bound(r,bds)
    type(rvec_t), dimension(:), intent(in) :: r
    type(rreg_t), intent(in), optional :: bds
    type(rreg_t) :: trg
    type(rvec_t) :: maxr,minr
    integer i
    integer, pointer :: d(:)

    if (present(bds)) then
       trg = bds
    else
       trg = geom_bounds
    end if

    if (.not.all(inside(trg,r))) then
       maxr = maxval(r)
       minr = minval(r)
       call conv_length_to_si(trg%min)
       call conv_length_to_si(trg%max)
       call conv_length_to_si(maxr)
       call conv_length_to_si(minr)
       write(std_err,*)'Particles out of geometry'
       write(std_err,*)'geom_bounds',trg
       write(std_err,*)'max(r)',maxr
       write(std_err,*)'min(r)',minr
       do i=1,0!size(r)       
          if (.not.inside(geom_bounds,r(i))) then
             write(std_err,*)'r',r(i)  
             call get_data(d,geom_loc,r(i))             ! Get the data
             write(std_err,*)'boundaries present',d
          end if
       end do

       !    call assert(all(inside(geom_bounds,r)), 'Particles out of geometry')
    end if
  end subroutine check_geom_bound

  ! Calculates particle wall contacts and interactions */ 
  subroutine calc_pw(part) 
    type(part_list_t),intent(inout) :: part
    integer i,j
    real(r_kind) :: max_ov,ov
    integer, pointer :: d(:)
    max_ov = 0
    do i=1,size(part%p)                ! For now just step through each particle
       call get_data(d, geom_loc,part%p(i)%mp%r)! Which boundary to check
       do j=1,size(d)                        ! How many boundaries in this region
          ov = calc_pb(part%p(i),d(j))       ! Check particle against wall k
          if (ov > max_ov) max_ov = ov
       end do
    end do
!    if(max_ov>col_props(PP_COLL)%max_ovlp) then
!       call conv_length_to_si(max_ov)
!       write(std_log,*)'Particles are overlapping wall too much',max_ov
!    end if
  end subroutine calc_pw 
  ! If it returns zero or a negative number this means no overlap is occuring
  ! If it returns a positive number this is the extent of the overlap
  ! By calling with constants aggressive inlining enables optimization
  function calc_pb(part,k) result(ov)
    type(particle_t),intent(inout) :: part
    integer, intent(in) :: k
    type(rvec_t) n
    real(r_kind) ::c1,c2
    real(r_kind) ov,r
    type(boundary_t) b 
    real(r_kind) n1x,n1y,t1x,t1y,d1,d2,d2a,d2b,r1,r2,mang
    type(rvec_t) n1
    !    type(rvec_t) p1,p2
    integer m 


    b = boundaries(k)
    ov = 0 ! overlap
    select case (b%type)              ! What sort of boundary
    case (XP_PLANE) ! x<s is solid
       ov = b%s(1)-part%mp%r%x+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(-1,0,0)),  ov)
    case (XM_PLANE) ! x>s is solid
       ov = part%mp%r%x-b%s(1)+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(1,0,0)),ov)
    case (YP_PLANE) ! y<s is solid
       ov = b%s(1)-part%mp%r%y+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(0,-1,0)), ov)
    case (YM_PLANE) ! y>s is solid
       ov = part%mp%r%y-b%s(1)+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(0,1,0)), ov)
#ifdef THREEDIM
    case (ZP_PLANE) ! z<s is solid
       ov = b%s(1)-part%mp%r%z+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(0,0,-1)), ov)
    case (ZM_PLANE) ! z>s is solid
       ov = part%mp%r%z-b%s(1)+P_RADIUS(part%mp)
       if (ov >0) call spring_force (part, rvec_t(ARG(0,0,1)), ov)
#endif 
    case (PLANE)
       ov = (b%v(1).dot.part%mp%r)- b%s(1)+P_RADIUS(part%mp)  ! Calculate overlap
       if (ov>0) call spring_force (part, b%v(1), ov)
    case (P_CYLINDER) ! interior of cylinder
       ! b%v(1) and b%v(2) are an orthonormal coordinate system
       ! let v(3) be orthogonal to b%v(1) and b%v(2) then centre line is
       ! b%v(1)*b%s(1) + lambda + v(3) 
       c1 = (b%v(1).dot.part%mp%r)-b%s(1) 
       c2 = b%v(2).dot.part%mp%r
       r = c1**2+c2**2
       if (r>(b%s(2)-P_RADIUS(part%mp))**2) then ! overlap has occured
          r = sqrt(r)
          n = (c1 * b%v(1) + c2 * b%v(2))/r
          ov = r - b%s(2) + P_RADIUS(part%mp)
          if (b%omega==0) then ! No angular velocity
             call spring_force (part, n, ov)
          else
             call spring_force (part, n, ov, n.x.b%w(1))
          end if
       end if
    case (M_CYLINDER) ! exterior of cylinder
       ! b%v(1) and b%v(2) are an orthonormal coordinate system
       c1 = b%s(1)-(b%v(1).dot.part%mp%r) 
       c2 = -b%v(2).dot.part%mp%r
       r = c1**2+c2**2
       if (r<(b%s(2)+P_RADIUS(part%mp))**2) then ! overlap has occured
          r = sqrt(r) 
          n = (c1 * b%v(1) + c2 * b%v(2))/r
          ov = b%s(2) - r + P_RADIUS(part%mp)
          if (b%omega==0) then ! No angular veclocity
             call spring_force (part, n, ov)
          else
             call spring_force (part, n, ov, n.x.b%w(1)) 
          end if
       end if
    case (R_CYLINDER) ! exterior of cylinder
       c1 = (b%v(1).dot.part%mp%r)-b%s(1) 
       c2 = b%v(2).dot.part%mp%r
       r = sqrt(c1**2+c2**2)
       if (abs(r-b%s(2))<b%s(4)+P_RADIUS(part%mp))then!Is a collision possible
          n = (c1 * b%v(1) + c2 * b%v(2))/r; ! Large circle normal
          m = floor((atan2(c2,c1)-b%theta)/b%aspace) ! Find nearest cylinders
          mang = (m+.5)*b%aspace + b%theta  ! Find mid angle
          n1x = cos(mang); n1y = sin(mang);
          t1x = -n1y; t1y = n1x;
          d1 = n1x*c1+n1y*c2-b%s(5);
          d2 = t1x*c1+t1y*c2; 
          d2a = d2-b%s(3); 
          d2b = d2+b%s(3); 
          r1 = sqrt(d1**2 + d2a**2);
          r2 = sqrt(d1**2 + d2b**2);
          if (r1<P_RADIUS(part%mp)+b%s(4)) then
             ov = P_RADIUS(part%mp)+b%s(4)-r1 
             n1 = -((d1*n1x + d2a*t1x)* b%v(1) + (d1*n1y + d2a*t1y)* b%v(2))/r1
             call spring_force (part, n1, ov, (n.x.b%w(1)) - (n1.x.b%w(2))) 
          end if
          if (r2<P_RADIUS(part%mp)+b%s(4)) then
             ov = P_RADIUS(part%mp)+b%s(4) - r2
             n1 = -((d1*n1x + d2b*t1x)* b%v(1) + (d1*n1y + d2b*t1y)* b%v(2))/r2
             call spring_force (part, n1, ov, (n.x.b%w(1)) - (n1.x.b%w(2))) 
          end if
       end if
!!$          p1 = (b%s(2)*cos(mang+.5*b%aspace)+b%s(1))*b%v(1)& 
!!$               + b%s(2)*sin(mang+.5*b%aspace)*b%v(2)
!!$
!!$          p2 = (b%s(2)*cos(mang-.5*b%aspace)+b%s(1))*b%v(1)& 
!!$               + b%s(2)*sin(mang-.5*b%aspace)*b%v(2)
!!$
!!$        call assert(r2-length(p2-part%mp%r),'r2-length(p2-part%mp%r)')
!!$       call assert(r1-length(p1-part%mp%r),'r1-length(p1-part%mp%r)')
!!$       call assert(length(r1*n1-p1 + part%mp%r),'n1 vec')
!!$       call assert(length(r2*n1-p2 + part%mp%r),'n2 vec')
    case (R_PLANE) ! exterior of cylinder
    end select
  end function calc_pb

  ! There is a contact of
  ! particle      : part 
  ! overlap       : ov 
  ! normal vector : norm
  ! with boundary : bn
  ! Dispatch to the correct routine to :
  !                 handle the interaction 
  !                 transform the particle
  !                 anhilate it
  subroutine resolve_contact(part, norm, ov, k)
    type(particle_t),intent(inout) :: part
    type(rvec_t), intent(in) :: norm
    real(r_kind), intent(in) :: ov
    integer, intent(in) :: k
    type(boundary_t) b
    b = boundaries(k)
    select case (b%sort)              ! What sort of boundary
    case(SPRING_BOUNDARY)           ! Ordinary springy boundary
       call spring_force(part, norm, ov)!, b%prop)
    case(TRANSFORM_BOUNDARY)       ! 
    end select
  end subroutine resolve_contact
  subroutine geom_info(fp)
    integer, intent(in) :: fp
    integer i;

    call conv_to_si(boundaries)
    call conv_to_si(geom_loc)

    write(fp,*) 'Geometry information'
    write(fp,*) 'Geometry name ', geom_name
    write(fp,*) 'Geometry type ', geom_type
    write(fp,*) 'Number of boundaries ', geom_num
    call write(fp,geom_loc,'Geometry locate_t')
    write(fp,*) 'Bounds : ', geom_bounds
    do i=1,geom_num 
       call info_boundary(fp, i)
    end do

    call conv_from_si(boundaries)
    call conv_from_si(geom_loc)


  end subroutine geom_info
  ! Display information abvout boundary b to fp
  subroutine info_boundary(fp, num)
    integer, intent(in) :: num
    integer, intent(in) :: fp
    integer i
    type(rreg_t), pointer :: r(:)
    write(fp,*)''
    write(fp,*)'Boundary',num
    call write_boundary(fp, boundaries(num))
    write(fp,*)'In regions'
    call find_data(r,geom_loc,num) ! Find the regions that have boundary num
    do i=1,size(r)
       write(fp,*)'region',r(i)
    end do
    deallocate(r)  
  end subroutine info_boundary
  ! Return the region corresponding to integer coordinates 

  ! Make some sanity checks on the loaded geometry
  subroutine check_geometry
    integer i
    do  i = 1,size(boundaries)
       call check_boundary(i)
    end do
  end subroutine check_geometry
  subroutine check_boundary(i)
    integer, intent(in) :: i
    type(boundary_t) b
    b = boundaries(i)
    select case (b%type)              ! What sort of boundary
    case (PLANE)
       call check_normal(b%v(1),'Plane')
    case (P_CYLINDER, M_CYLINDER) ! interior of cylinder
       call check_normal(b%v(1),'Cylinder vec 1')
       call check_normal(b%v(2),'Cylinder vec 2')
       call assert(b%v(1).dot.b%v(2),'Cylinder vectors not orthogonal')
    end select
  end subroutine check_boundary
  ! check a vector is a unit vector
  ! if it i snot print a warning message and normalise it
  subroutine check_normal(r,s)
    type(rvec_t),intent(inout) :: r
    character(len=*), intent(in) :: s
    real(r_kind) n
    n = length(r)
    if (abs(n-1)>10*epsilon(n)) then
       call warning_msg('The norm of the vector '//s//' in geometry.f90 is not one')
       r = r/n
    end if
  end subroutine check_normal
end module geometry
