#include "dimension.h"
#include "radius.h"
! Module that allocates a grid and provided grid operating functions
! Note that this code is highly nonrentrant and uses numerous static variables 
! Functions are 
! grid_to_point
! init_grid 
! end_grid
! coord_to_grid
! Calculate which grid cells each particle is in     
! Convert from a grid number to coordinates
! Uses a symmetric coordinate system 
! Indexing the grid points  
! Needs to be a macro otherwise we can't inline

module grid
  use precision
  use sysdep_io
  use program_units
  use vector
  implicit none
  !private SR2201BUG

  ! Public procedures
  public :: init_grid        ! Set the grid up for a particular region
  public :: add_to_grid      ! Add particles to the grid
  public :: check_grid_point ! Return particles at a grid point
  public :: zero_grid        ! Clear grid points
  public :: push             ! Push particle onto a grid point
  public :: grid_info        ! Print information about the current grid
  public :: grid_bounds      ! Return the current grid bounds
  public :: coord_to_grid    ! Convert a point to a grid coordinate
  public :: offset_grid      ! Offset the grid by a vector
  public :: grid_contains    ! Does the grid contain a certain region
  public :: end_grid         ! Finish of the grid and print usage stats
  public :: adj_part         ! return adjacent particles
  public :: h_adj_part       ! return half of adjacent particles

  ! Public types
  public :: grid_element, gr_part_list

  real(r_kind), parameter :: G_SIZE = 2*max_radius ! Size of grid squares
  real(r_kind), parameter :: EG_SIZE = RADIUS_RATIO+1

  integer, parameter :: GR_ALL_N = 3**DIM ! Total number of all the squares
  integer, parameter :: GR_ADJ_N = (3**DIM-1)/2 ! half adjacent squares
  integer, parameter :: GR_N_MAX = 2*(RADIUS_RATIO+1)**DIM 

  integer, parameter :: GR_SPARE_SIZE = 1 ! Allow one particle in all directions

  ! Memory allocation policy No code at present
  integer, parameter :: GR_E_SIZE =  (GR_N_MAX+1)*4 ! Number of bytes per grid element
  integer, parameter :: MAX_MEM  = 128*1024*1024  ! maximum amount of memory to use
  integer, parameter :: MAX_SIZE =  MAX_MEM/GR_E_SIZE  ! maximum amount of memory to use
  logical, parameter :: MAX_ALLOC = .false. ! Allocate the maximum amount of memory

  ! Collect some statistics
  integer :: count_grid_init = 0 ! How many times was grid_init called
  integer :: count_mem_alloc = 0 ! How many times was grid memory allocated
  integer :: max_grid_size   = 0 ! How big is the largest allocated grid
  type :: grid_element       ! Type for each grid element
     integer pn(GR_N_MAX) ! Which particles are in the grid cube
     integer n            ! How many particles in a grid cube 
  end type grid_element
    
  type, public ::  grid_type         ! type for a whole grid
     type(grid_element), pointer ::g(:) ! array for grid squares
     type(grid_element), pointer ::gt(:) ! allocated memory
     type(ivec_t) sz        ! dimension of grid in point sizes
     integer size          ! size of the grid
     integer off           ! offset used for addressing the grid
     integer h_adj(GR_ADJ_N) ! offsets for half the adjacent grid squares
     integer adj(GR_ALL_N) ! offsets for half the adjacent grid squares
     integer recalc       ! Estimated program steps until a recalc is needed
     type(rreg_t) edge    ! region grid is good for
     integer f_type       ! Type of face
     type(rvec_t) r3_off ! Extra offset to allow for dynamic grid movement
  end type grid_type
  
  type ::  gr_part_list
     integer n       ! Number of particles
     integer pn (GR_ALL_N*GR_N_MAX) ! Array of particle numbers
  end type gr_part_list
  
  interface assignment(=)
     module procedure ge_equ_ge
  end interface
  interface coord_to_grid
     module  procedure s_coord_to_grid, a_coord_to_grid
  end interface
  interface grid_to_point
     module procedure s_grid_to_point, a_grid_to_point
  end interface
  interface add_to_grid
     module procedure s_add_to_grid, a_add_to_grid, ai_add_to_grid
  end interface
  interface push
     module procedure push_ge_pl, push_pl_ge, push_pl_p,push_ge_p
  end interface
  type(grid_type), save :: gd ! global variable for the grid data 


  logical, private, parameter :: debug_grid = .true. !debugging messages
contains ! main functions
  ! Close up the grid module and print information about usage stats to fp
  ! only print information on root processor
  subroutine end_grid(fp)
    use mpi
    integer, intent(in) :: fp
    integer,dimension(2) :: data_in, data
    integer :: max_sz
    data_in(1) = count_grid_init
    data_in(2) = count_mem_alloc
    call fmpi_reduce(data_in, data, FMPI_SUM)
    call fmpi_reduce(max_grid_size, max_sz, FMPI_MAX)
    if (fmpi_root) then
       write(fp,*)'Grid module information'
       write(fp,*)'Grid init called',data(1),'times.'
       write(fp,*)'Memory allocated',data(2),'times.'
       write(fp,*)'Largest individual grid', max_sz
    end if
    deallocate(gd%gt)
  end subroutine end_grid
  subroutine offset_grid(r)
    type(rvec_t), intent(in) :: r
    gd%r3_off = gd%r3_off + r
    gd%edge%min =  gd%edge%min + r   ! region grid is good for
    gd%edge%max =  gd%edge%max + r   ! region grid is good for
  end subroutine offset_grid
  ! Write all the information out about the grid to a file, possibly stdout
  subroutine grid_info(fp)
    integer, intent(in) :: fp
    write(fp,*) 'Grid information'
    if (associated(gd%g)) then
       write(fp,*) 'Grid elements have been allocated : ', size(gd%g) 
       write(fp,*) 'The grid contains particle : ', sum(gd%g%n)
    else
       write(fp,*) 'Grid has not been allocated'
    end if
    write(fp,*) 'Size : ', gd%sz, ', total ', gd%size
    write(fp,*) 'Offset : ', gd%off
    write(fp,*) 'R3 offset : ', gd%r3_off 
    write(fp,*) 'Edge : ', gd%edge 
  end subroutine grid_info
  ! Check whether the grid contains a certain region
  ! Not that the upper inequality is strict and the lower isn't
  function grid_contains(rg)
    type(rreg_t), intent(in) :: rg
    logical grid_contains
    grid_contains = gd%edge%min<=rg%min .and. rg%max < gd%edge%max
  end function grid_contains
  ! create a grid large enough to handle particles 
  ! within the specified region
  ! maybe initialized with a zero region corresponding to one grid square
  ! maybe called many times an should efficiently reuse the memory
  ! should be allowed to assume that only newly allocated memory needs to be
  ! zeroed  
  subroutine init_grid(rga)
    type(rreg_t), optional, intent(in) :: rga ! real(r_kind) region
    type(ireg_t) :: outer ! integer region in G_SIZE units
    type(rreg_t) :: rg 

    count_grid_init = count_grid_init + 1; ! Count invocations
    
    if (present(rga)) then; rg = rga
    else; rg = null_rreg
    end if
    if (debug_grid) write(std_log,*)'Init grid called',rg
    ! Calculate coordinates grid is good for    
    
    ! Start of with no offset
    gd%r3_off = null_rvec ! No real position offset%off

    ! Check all the coordinates are somewhere sensible
    ! This generaly occurs beacuse of taking max and min of zero length arrays
    if (.not.(rg%max>=rg%min)) then
       rg=null_rreg
       write(std_log,*)'Coordinates are negative volume region. All set to 0'
    end if
    ! Calculate the coordinates of the particles in grid coordinates
    outer%max = scalari_to_ivec(1)+floor((rg%max-gd%r3_off)/G_SIZE)
    outer%min = floor((rg%min-gd%r3_off)/G_SIZE)
    ! Calculate coordinates of points on the edge
    gd%edge%min = G_SIZE * outer%min ! This is the region the grid is good for
    gd%edge%max = G_SIZE * outer%max ! This is the region the grid is good for
    ! so the grid is good for all points
    ! gd%edge%min <= x < gd%edge%max

    ! allow an extra border around the grid for overlap calulations
    outer = grow(outer, GR_SPARE_SIZE) !Allow one spare cell
    ! calculate the grid size how many steps in each directions
    gd%sz = outer%max-outer%min
    gd%size = volume(outer) ! This is the size of the new grid  
    if (debug_grid) write(std_log,*)"Resizing grid to region",&
         gd%edge,', size',gd%size
    if (associated(gd%gt)) then 
       if ((size(gd%gt)<gd%size) .or. (size(gd%gt)>2*gd%size)) then
          deallocate(gd%gt)
       end if
    end if

    if (.not.associated(gd%gt)) then 
       allocate(gd%gt(gd%size))
       max_grid_size = max(max_grid_size,gd%size)
       count_mem_alloc = count_mem_alloc + 1
       if (debug_grid) write(std_log,*)"Grid reallocating memory", &
            gd%size*32./1024./1024.,"MB"
       gd%gt%n=0  ! Clear the counts for all the grid squares
    end if

    gd%g => gd%gt(1:gd%size) ! Point grid squares
    call assert(all(gd%g%n==0),'Not all grid squares are empty')
      
    ! Now calculate grid offsets for half of adjacent squares 
    gd%off = 0
#ifdef THREEDIM
    gd%h_adj = grid_to_point((/&
         ivec_t(0,-1, 1),ivec_t(0, 0, 1),ivec_t(0, 1, 1),ivec_t(0, 1, 0),&
         ivec_t(1,-1,-1),ivec_t(1,-1, 0),ivec_t(1,-1, 1),&
         ivec_t(1, 0,-1),ivec_t(1, 0, 0),ivec_t(1, 0, 1),&
         ivec_t(1, 1,-1),ivec_t(1, 1, 0),ivec_t(1, 1, 1)/))
    gd%adj = grid_to_point((/&
         ivec_t(-1,-1,-1),ivec_t(-1,-1, 0),ivec_t(-1,-1, 1),&
         ivec_t(-1, 0,-1),ivec_t(-1, 0, 0),ivec_t(-1, 0, 1),&
         ivec_t(-1, 1,-1),ivec_t(-1, 1, 0),ivec_t(-1, 1, 1),&
         ivec_t( 0,-1,-1),ivec_t( 0,-1, 0),ivec_t( 0,-1, 1),&
         ivec_t( 0, 0,-1),ivec_t( 0, 0, 0),ivec_t( 0, 0, 1),&
         ivec_t( 0, 1,-1),ivec_t( 0, 1, 0),ivec_t( 0, 1, 1),&
         ivec_t( 1,-1,-1),ivec_t( 1,-1, 0),ivec_t( 1,-1, 1),&
         ivec_t( 1, 0,-1),ivec_t( 1, 0, 0),ivec_t( 1, 0, 1),&
         ivec_t( 1, 1,-1),ivec_t( 1, 1, 0),ivec_t( 1, 1, 1)/))
#else
    gd%h_adj = grid_to_point(&
         (/ivec_t(-1,1),ivec_t(0,1),ivec_t(1,1),ivec_t(1,0)/))
    gd%adj = grid_to_point((/& 
         ivec_t(-1,-1),ivec_t(-1, 0),ivec_t(-1, 1), &
         ivec_t( 0,-1),ivec_t( 0, 0),ivec_t( 0, 1), &
         ivec_t( 1,-1),ivec_t( 1, 0),ivec_t( 1, 1)/))
#endif
    ! Now set the offset coordinate for the smallest point
    gd%off = 1 - coord_to_grid(G_SIZE*outer%min)  ! calculate the grid offset 

  end subroutine

  function grid_bounds() 
    type(rreg_t) :: grid_bounds
    grid_bounds = gd%edge
  end function grid_bounds
 
    
    ! add a particle to the grid at point r and return the grid point
    subroutine s_add_to_grid(gn, r, pn)
      type(rvec_t), intent(in) 	::r            ! particle position
      integer, intent(in) 	::pn           ! particle number
      integer, intent(out)	::gn           ! grid number
      gn = coord_to_grid(r)
      call push(gd%g(gn),pn)
    end subroutine s_add_to_grid
    subroutine a_add_to_grid(gn, r)
      type(rvec_t), intent(in):: r(:)
      integer, intent(out) :: gn(size(r))
      integer i
     do i = 1,size(r,1)
        call s_add_to_grid(gn(i), r(i), i)
      end do
    end subroutine a_add_to_grid
    subroutine ai_add_to_grid(gn) ! Add all the particles 
      integer, intent(in) :: gn(:)
      integer i
      do i = 1,size(gn,1)
         call push(gd%g(gn(i)),i)
      end do
    end subroutine ai_add_to_grid
    subroutine zero_grid(gn) ! Clear the grid squares specified by gn
      integer,intent(in) :: gn(:)
      gd%g(gn)%n=0
    end subroutine zero_grid
    
    ! return a list of the particles in a grid point
    subroutine check_grid_point(pn, gn)
      type(grid_element), intent(out)::pn       ! list of the particles
      integer,  intent(in)           ::gn       ! node number
      pn = gd%g(gn)                           ! = is overloaded
    end subroutine check_grid_point
    
    ! return a list of the particles in half the surrounding grid points
    subroutine h_adj_part(pn, gn)
      type(gr_part_list), intent(out)::pn       ! list of the particles
      integer, intent(in)            ::gn       ! node number
      integer i
      pn%n = 0 
      do i = 1, GR_ADJ_N ! look through half adjacent grid squares
         call push(pn,gd%g(gn+gd%adj(i)))!copy all the particles
      end do
    end subroutine h_adj_part
    subroutine adj_part(pn, gn)
      type(gr_part_list), intent(out)::pn       ! list of the particles
      integer, intent(in)            ::gn       ! node number
      integer i
      pn%n = 0 
      do i = 1, GR_ALL_N ! look through all adjacent grid squares
         call push(pn,gd%g(gn+gd%adj(i)))!copy all the particles
      end do
    end subroutine adj_part

    function s_coord_to_grid (r) !converting coordinates to grid points
      integer  s_coord_to_grid
      type(rvec_t), intent(in) :: r
      s_coord_to_grid = grid_to_point(floor((r-gd%r3_off)/G_SIZE))
    end function s_coord_to_grid
    function a_coord_to_grid (r) !converting coordinates to grid points
      type(rvec_t), intent(in) :: r(:)
      integer  :: a_coord_to_grid(size(r))
      if (gd%r3_off == null_rvec) then
         a_coord_to_grid = grid_to_point(floor(r/G_SIZE))
      else
         a_coord_to_grid = grid_to_point(floor((r-gd%r3_off)/G_SIZE))
      end if
    end function a_coord_to_grid
    
    function s_grid_to_point(a) result(p)
      type(ivec_t), intent(in)  :: a
      integer p
      p = gd%off+(a%x*gd%sz%y+a%y) CMD3(*gd%sz%z+a%z)
    end function s_grid_to_point
    function a_grid_to_point(a) result(p)
      type(ivec_t), intent(in)  :: a(:)
      integer p(size(a))
      p = gd%off+(a%x*gd%sz%y+a%y)CMD3(*gd%sz%z+a%z)
    end function a_grid_to_point
 ! first start off with overloaded functions
  subroutine ge_equ_ge(a,b) ! copy n particles from b to a
    type(grid_element), intent(out) :: a
    type(grid_element), intent(in) :: b
    a%n = b%n;      
    a%pn(1:b%n) = b%pn(1:b%n)
  end subroutine ge_equ_ge
  subroutine push_ge_pl(a,b) ! push from a grid b to a list of particle nos
    type(gr_part_list), intent(inout) :: a
    type(grid_element), intent(in) :: b
    a%n = a%n + b%n; a%pn((a%n+1-b%n):a%n) = b%pn(1:b%n);     
  end subroutine push_ge_pl
   subroutine push_pl_ge(a,b)
    type(grid_element), intent(inout) :: a
    type(gr_part_list), intent(in) :: b
    a%n = a%n + b%n; a%pn((a%n+1-b%n):a%n) = b%pn(1:b%n);     
  end subroutine push_pl_ge
  subroutine push_pl_p(a,b) ! add a particle to a list
     type(gr_part_list), intent(inout) :: a
     integer, intent(in) :: b
     a%n = a%n + 1; a%pn(a%n) = b
   end subroutine push_pl_p
  subroutine push_ge_p(a,b) ! add a particle to a grid elements
     type(grid_element), intent(inout) :: a
     integer, intent(in) :: b
     a%n = a%n + 1; a%pn(a%n) = b
   end subroutine push_ge_p
end module grid
