#include "dimension.h"
! Fortran file for reading and writing initial state
! An initial state can either consist of a list of positions and velocities
! or a specification string for example create a cuboid of size 2x3x4 
! of particles 
!g
! data types are
!
module initial_state
  use precision
  use vector
  use matrix
  use program_units 
  use particle
  use sysdep_io 
  use collision
  implicit none
  !private SR2201BUG

  ! Functions to save and read a state
  public :: read_initial_state, info_initial_state

  ! Some have sensible default values others should generate an error
  ! if absent
  logical           :: random_lin_vel = .false.    ! Give random linear velocities
  logical           :: random_ang_vel = .false.    ! Give random angular velocities
  real(r_kind)      :: lin_vel_max = 0             ! Maximum random speed
  real(r_kind)      :: ang_vel_max = 0             ! Maximum random speed
  character(len=10) :: initial_type = 'No type   ' ! Type of initial data
  integer           :: program_step = 0            ! Program step to start from
  real(r_kind)      :: program_time = 0            ! Program time
  integer           :: particle_num = -1           ! Number of particles
  type(ivec_t)       :: cube_size   ! size of cube of particles
  real(r_kind)      :: packing_sep = 1             ! packing separation in 2*RADIUS units
  type(rvec_t)       :: origin         ! offset for the particle positions
  type(rvec_t)       :: vel_offset     !offset for the particle velocities
  type(rang_t)       :: avel_offset! offset for the particle angular velocities
  type(rang_t)       :: rotation       ! rotation of state, before offset
  character(len=FILE_NAME_LEN) :: file_name

  namelist /particle_state/ &
       initial_type, &
       file_name, &
       particle_num, &
       program_step, &
       program_time, &  
       random_lin_vel, &
       random_ang_vel, &
       lin_vel_max, &
       ang_vel_max, &
       cube_size, &
       packing_sep, &
       vel_offset, &
       avel_offset, &
       origin, &
       rotation

! Set this to print extra debugging info from this file 
  logical, parameter, private :: is_debug = .false.

contains  ! Print information about an initial state file
  subroutine info_initial_state(fp, outfp)
    integer outfp  ! Where to print info
    integer fp     ! Unit number for the file
    write(outfp,*) 'Particle initial state info'

    cube_size = null_ivec   ! size of cube of particles
    origin=null_rvec         ! offset for the particle positions
    vel_offset=null_rvec     !offset for the particle velocities
    avel_offset=null_rang! offset for the particle angular velocities
    rotation=null_rang       ! rotation of state, before offset
 
    read(fp,nml=particle_state)
    write(outfp,*)'Number of particles : ', particle_num
    write(outfp,*)'Program step : ', program_step
    write(outfp,*)'Program time : ', program_time
    if (random_lin_vel) then
       write(outfp,*)'Random linear velocities max is : ', lin_vel_max
    else
       write(outfp,*)'No random linear velocities'
    end if
    if (random_ang_vel) then
       write(outfp,*)'Random angular velocities max is : ', ang_vel_max
    else
       write(outfp,*)'No random angular velocities'
    end if
    select case (trim(initial_type))
    case('data')
       write(outfp,*)'Initial data is in file ', file_name
    case('cuboid')
       write(outfp,*)'Initial data is a cube of dimensions : ' , cube_size
       write(outfp,*)'Packing sepaation is : ', packing_sep
    case default
       write(outfp,*)'Unimplimented initial data : ', initial_type
    end select
    if (length(rotation) /=0) then
       write(outfp,*)'Rotation vector is : ', rotation   
    else
       write(outfp,*)'No rotation specified' 
    end if
    ! Offset positions
    if (length(origin) /=0) then
       write(outfp,*)'Offset origin is at : ', origin
    else 
       write(outfp,*)'No offset origin specified' 
    end if
     if (length(vel_offset) /=0) then
       write(outfp,*)'Velocity offset is : ', vel_offset
    else 
       write(outfp,*)'No velocity offset specified' 
    end if
     if (length(avel_offset) /=0) then
       write(outfp,*)'Angular velocity offset is : ', avel_offset
    else 
       write(outfp,*)'No angular velocity offset specified' 
    end if
  end subroutine info_initial_state

  ! Read an initial state
  subroutine read_initial_state(fp, part)
    integer, intent(in) :: fp ! File pointer
    type(min_part_list_t),intent(out) :: part
    type(rmat_t) :: rot_m ! Rotate matrix
    logical :: have_part_props = .false.

    read(fp,nml=particle_state)

    call alloc(part,particle_num,particle_num)
    select case (trim(initial_type))
    case('data')
       call initial_conditions_data(trim(file_name), part)
       have_part_props = .true. ! This loads the particle properties
    case('cuboid')
       call initial_conditions_cuboid(part)
    case('closepack')
       call error_halt('Unimplimented initial data ' // trim(initial_type))
    case('random')
       call error_halt('Unimplimented initial data ' // trim(initial_type))
    case default
       call error_halt('Unknown initial data : ' // trim(initial_type)) 
    end select
    close(fp) ! No longer using the file
    ! rotate all initial vectors
    if (length(rotation) /=0 ) then
       rot_m = rotate_matrix(rotation)
       part%p%r = rot_m * part%p%r 
       part%p%v = rot_m * part%p%v 
       part%p%w = rot_m * part%p%w 
    end if
    ! Offset positions
    if (length(origin) /=0 ) then
       part%p%r = part%p%r + origin
    end if
    if (length(vel_offset) /=0 ) then
       part%p%v = part%p%v + vel_offset
    end if
    if (length(avel_offset) /=0 ) then
       part%p%w = part%p%w + avel_offset
    end if
    if (random_lin_vel) then
       call random_unit_vector(part%p%v)
       part%p%v = lin_vel_max * part%p%v
    end if
    if (random_ang_vel) then
       call random_unit_vector(part%p%w)
       part%p%w = ang_vel_max * part%p%w
    end if
    if (is_debug) then
       write(std_log,*)'initial positions =',part%p%r 
       write(std_log,*)'initial velocities =',part%p%v 
    end if
    call conv_from_si(part%p) ! Convert to program units
    if (.not. have_part_props) then
       call set_part_props(part%p)!Set the particle types collision props etc.
    end if
  end subroutine read_initial_state

  ! Read an initial state 
  subroutine initial_conditions_data(fname, part)
    character(len=*), intent(in) :: fname
    type(min_part_list_t), intent(inout) :: part
    integer :: fp

    call assign_lun(fp) ! Get a unit number
    call info_msg('Reading initial conditions from ' // fname)
    open(fp,file=fname, status='old', action='read')
    read(fp,*)part%p
!!$    call read_part_state(fp,part%p)
!!$    call read_part_props(fp,part%p)
    close(fp)  ! Close the file 
  end subroutine initial_conditions_data

  ! Create a cuboid  in SI units
  subroutine initial_conditions_cuboid(part)
    type(min_part_list_t), intent(inout) :: part
    integer  :: n 
    integer :: ARG(i,j,k)
    real(r_kind) sep
    type(rvec_t)  off
    n = size(part%p)
    call info_msg('Making cube of particles ')
!SR2201BUG    call assert(volume(cube_size)>=size(part%p), &
!         'Number of particles is greater than cube size')
    off = -real(0.5,r_kind)*(cube_size + scalar_to_ivec(1))
    n = 0
    sep = 2*max_radius*packing_sep 
    call conv_length_to_si(sep)
    outer:do i = 1, cube_size%x
       do j = 1, cube_size%y
#ifdef THREEDIM
          do k = 1,cube_size%z
#endif
             n = n + 1
             part%p(n)%pn = n
             part%p(n)%r = sep*(ivec_t(ARG(i,j,k))+off)
             if (n==size(part%p)) exit outer
          CMD3(end do)
       end do
    end do outer
    part%p%v = null_rvec
    part%p%w = null_rang
  end subroutine initial_conditions_cuboid
end module initial_state
