#include "dimension.h"
! This module contains functions for associating integer data with
! region of space and then providing lookup to this data
! find_data ! find data returns an array of  regions contains integers
! an mpi routine for broadcasting this data is also available
!
! locate_t the type definition for objects in this module
! allocate       create a locate object
! deallocate     deletea a locate object 
! lookup         find the region of a point
! get_data       find the data corresponding to a point
! find_data      find which regions contain certain data
! write          write a locate object to a file
! read           read a locate object from a file
! conv_from_si   convert a locate object from SI units
! conv_to_si     convert a locate object to SI units
! comms_bcast    broadcast a locate object to all processes
! push           push a piece of data onto a node
! pop            pop a piece of data from a node
module locate
  use precision
  use vector
  use program_units
  use sysdep_io
  implicit none
  !private SR2201BUG
  public :: lookup, get_data, test_locate_module, find_data, comms_bcast
  public :: write, read, allocate, deallocate, conv_from_si, conv_to_si
  public :: push
  interface comms_bcast
     module procedure comms_bcast_locate_t_s
  end interface
  interface find_data
     module procedure find_data_s
  end interface
  interface write
     module procedure write_locate_t
  end interface
  interface read
     module procedure read_locate_t
  end interface
  interface allocate
     module procedure allocate_locate_t
  end interface
  interface deallocate
     module procedure deallocate_locate_t
  end interface

  interface conv_to_si
     module procedure conv_locate_to_si
  end interface
  interface conv_from_si
     module procedure conv_locate_from_si
  end interface
  interface lookup
     module procedure s_lookup, a_lookup, sv_lookup, av_lookup
  end interface
  interface get_data
     module procedure get_data_sp, get_data_s
  end interface
  interface push
     module procedure ds_i_push_lt, da_i_push_lt
  end interface

  type, public :: locate_t
     logical               :: initialized = .false. ! Has it been allocated
     type(rreg_t)          :: rg       ! region containing data
     type(ivec_t)          :: nc       ! number of cuts along each axis
     integer               :: sz       ! sum(nc)
     real(r_kind), pointer :: allc(:)  ! array of all the cutsx1
     real(r_kind), pointer :: xc(:)      ! cuts along x direction
     real(r_kind), pointer :: yc(:)      ! cuts along y direction
#ifdef THREEDIM
     real(r_kind), pointer :: zc(:)      ! cuts along z direction
#endif
     integer               :: max_n! largest number of elements in a grid square
     integer, pointer      :: n(ARG(:,:,:))   ! How many elements in a grid square
     integer, pointer      :: d(ARG(:,:,:),:) ! data elements  in a grid sqaure
  end type locate_t
  logical, parameter, private :: debug_locate = .false. 
contains
  subroutine comms_bcast_locate_t_s(lt)
    use mpi
    type(locate_t), intent(inout) :: lt
    call fmpi_bcast(lt%nc)
    call fmpi_bcast(lt%rg)
    call fmpi_bcast(lt%max_n)
    if (.not.fmpi_is_root()) call allocate(lt,lt%nc,lt%max_n)
    call fmpi_bcast(lt%allc)
    call fmpi_bcast(lt%n)
    call fmpi_bcast(lt%d)
  end subroutine comms_bcast_locate_t_s
  subroutine write_locate_t(fp,lt,str)
    integer, intent(in) :: fp
    type(locate_t), intent(in) :: lt
    character(len=*), optional :: str
    integer ARG(i,j,k),n
    if (.not.present(str)) then
       write(fp,*)lt%xc
       write(fp,*)lt%yc
       CMD3(write(fp,*)lt%zc)
       write(fp,*)lt%n   
       write(fp,*)lt%d
    else
       write(fp,*)'Locate type ',str
       write(fp,*)'max n',lt%max_n
       write(fp,*)'ncx',lt%nc%x,'x cuts',lt%xc
       write(fp,*)'ncy',lt%nc%y,'y cuts',lt%yc
#ifdef THREEDIM
       write(fp,*)'ncz',lt%nc%z,'z cuts',lt%zc
#endif
       do i=0,lt%nc%x
          do j=0,lt%nc%y
#ifdef THREEDIM
              do k=0,lt%nc%z
#endif
                n = lt%n(ARG(i,j,k))
                write(fp,*)'n =',n,'(ARG(i,j,k)) =',ARG(i,j,k),&
                     'data = ',lt%d(ARG(i,j,k),1:n)
             end do
          CMD3(end do)
       end do
    end if
  end subroutine write_locate_t  
  subroutine read_locate_t(fp,lt)
    integer, intent(in) :: fp
    type(locate_t), intent(inout) ::lt
    integer i    
    if (lt%initialized) call deallocate_locate_t(lt)
    read(fp,*)lt%nc
    call allocate(lt,lt%nc)
    read(fp,*)lt%xc
    read(fp,*)lt%yc
    CMD3(read(fp,*)lt%zc)
    read(fp,*)lt%n   
    lt%max_n = maxval(lt%n)
    allocate(lt%d(ARG(0:lt%nc%x,0:lt%nc%y,0:lt%nc%z),1:lt%max_n))
    do i=1,lt%max_n
       read(fp,*)lt%d(ARG(:,:,:),i)
    end do
    lt%rg%min=scalar_to_rvec(-999999)
    lt%rg%max=scalar_to_rvec(-999999)
  end subroutine read_locate_t
  subroutine allocate_locate_t(lt,s,max_n)
    type(locate_t), intent(out) :: lt
    type(ivec_t), intent(in) :: s
    integer, intent(in), optional :: max_n
    if (debug_locate) write(std_err,*)'Allocating locate type size',lt%initialized
!SR2201BUG    allocate(lt%allc(s%x+s%y CMD3(+s%z)))
    if (lt%initialized) call deallocate_locate_t(lt)
    lt%initialized = .true.
    allocate(lt%allc(s%x+s%y CMD3(+s%z) + 1))
    lt%nc = s
    lt%xc => lt%allc(1:s%x)
    lt%yc => lt%allc((s%x+1):(s%x+s%y))
    CMD3(lt%zc => lt%allc((s%x+s%y+1):(s%x+s%y+s%z)))
    allocate(lt%n(ARG(0:s%x,0:s%y,0:s%z)))
    if (present(max_n)) then
       lt%max_n = max_n
       allocate(lt%d(ARG(0:s%x,0:s%y,0:s%z),1:max_n))
    end if
  end subroutine allocate_locate_t
  subroutine deallocate_locate_t(lt)
    type(locate_t), intent(inout) :: lt
    lt%initialized = .false.
    deallocate(lt%allc)
    nullify(lt%xc)
    nullify(lt%yc)
    CMD3(nullify(lt%zc))
    deallocate(lt%n)
    deallocate(lt%d)
  end subroutine deallocate_locate_t
  subroutine conv_locate_to_si(lt)
    type(locate_t), intent(inout) :: lt
    call conv_length_to_si(lt%allc) ! Convert units
  end subroutine conv_locate_to_si
  subroutine conv_locate_from_si(lt)
    type(locate_t), intent(inout) :: lt
    call conv_length_from_si(lt%allc) ! Convert units
  end subroutine conv_locate_from_si
  ! lookup the position x in the table lt and return 
  ! the data objects in the pointer d
  subroutine get_data_sp(d,lt,r)
    integer, pointer :: d(:)
    type(locate_t), intent(in) :: lt
    type(rvec_t), intent(in)      :: r
    type(ivec_t) :: hc
    hc = lookup(r,ARG(lt%xc,lt%yc,lt%zc))   ! Get coordinates
    d => lt%d(ARG(hc%x,hc%y,hc%z),1:lt%n(ARG(hc%x,hc%y,hc%z)))     ! Return data
  end subroutine get_data_sp
  subroutine get_data_s(n,d,lt,r)
    integer, intent(out) :: n
    integer, intent(out) :: d(:)
    type(locate_t), intent(in) :: lt
    type(rvec_t), intent(in)      :: r
    type(ivec_t) :: hc

    hc = lookup(r,ARG(lt%xc,lt%yc,lt%zc))   ! Get coordinates
    n = lt%n(ARG(hc%x,hc%y,hc%z))           ! Number of data points
    !   call assert(n<=lt%max_n,'n is not set in locate type')
    d = lt%d(ARG(hc%x,hc%y,hc%z),1:n)       ! Return data
  end subroutine get_data_s

  ! returns g such that r(g)<=x<r(g+1) 
  ! g maybe 0 indicating x<r(1)
  ! g length(r) indicating x>=r(length(r))
  ! if the initial guess is present check that region and the two surrounding regions 
  ! first
  ! at any point l<=g<u
  ! If an initial guess is present then x(gs-1:gs+2) must all exist
  function s_lookup(p,x,gs) result(l)
    real(r_kind), intent(in) :: p
    real(r_kind), intent(in) :: x(:)
    integer, optional, intent(in) :: gs ! Initial guess for coordinate
    integer :: g
    integer :: l
    integer :: u
    if (present(gs)) then
       if (p>=x(gs)) then
          if (p<x(gs+1)) then
             l = gs
             return
          else if (p<x(gs+2)) then
             l = gs+1
             return
          else
             l = gs+2
             u = size(x)+1
          end if
       else if (p>=x(gs-1)) then
          l = gs-1
          return
       else
          l = 0
          u = gs-1
       end if
    else
       l = 0
       u = size(x)+1
    end if
    do
       if (u==l+1) exit
       g = (u+l)/2
       if (p>=x(g)) then
          l = g
       else 
          u = g
       end if
    end do
  end function s_lookup
  function a_lookup(p,x,gs) result(g)
    real(r_kind), intent(in) :: p(:)
    real(r_kind), intent(in) :: x(:)
    integer, optional, intent(in) :: gs(size(p))
    integer :: g(size(p))
    integer :: i;
    if (present(gs)) then
       do i = 1,size(p)
          g(i) = lookup(p(i),x,gs(i))
       end do
    else
       do i = 1,size(p)
          g(i) = lookup(p(i),x)
       end do
    end if

  end function a_lookup
  function sv_lookup(p,ARG(x,y,z),gs) result(g)
    type(rvec_t), intent(in) :: p
    real(r_kind), intent(in) :: ARG(x(:),y(:),z(:))
    type(ivec_t), optional, intent(in) :: gs
    type(ivec_t) :: g
    if (present(gs)) then
       g%x = lookup(p%x,x,gs%x)
       g%y = lookup(p%y,y,gs%y)
       CMD3(g%z = lookup(p%z,z,gs%z))
    else
       g%x = lookup(p%x,x)
       g%y = lookup(p%y,y)
       CMD3(g%z = lookup(p%z,z))
    end if

  end function sv_lookup
  function av_lookup(p,ARG(x,y,z),gs) result(g)
    type(rvec_t), intent(in) :: p(:)
    real(r_kind), intent(in) :: ARG(x(:),y(:),z(:))
    type(ivec_t), optional, intent(in) :: gs(size(p))
    type(ivec_t) :: g(size(p))
    if (present(gs)) then
       g%x = lookup(p%x,x,gs%x)
       g%y = lookup(p%y,y,gs%y)
       CMD3(g%z = lookup(p%z,z,gs%z))
    else
       g%x = lookup(p%x,x)
       g%y = lookup(p%y,y)
       CMD3(g%z = lookup(p%z,z))
    end if
  end function av_lookup
  ! tests all the procedures intrinsic to the locate module
  ! defines 

  subroutine test_locate_module
#ifdef THREEDIM
    real(r_kind) :: x(5) = (/2.1, 2.2, 3., 4., 5./)
    real(r_kind) :: y(4) = (/2.1, 2.2, 2.3, 2.4/)
    real(r_kind) :: z(1) = (/0./)
    real(r_kind) :: p(9) = (/ -5., 0., .3, 2.1, 2.15, 3., 3.5, 5., 6./)
    integer      :: g(9) = (/ 0,  0,  0,   1, 1,    3, 3,   5, 5/)
    integer      :: a(size(g))
    type(rvec_t)    :: pv(2) = (/rvec_t(2.25,2.35,.1),rvec_t(4.,5,-.1)/)
    type(ivec_t)    :: gv(2) = (/ivec_t(2,3,1),ivec_t(4,4,0)/)
    type(ivec_t)    :: av(2)
    call info_msg('Testing locate module',debug_locate)
    a = lookup(p,x)
    av = lookup(pv,x,y,z)
    if (any(a/=g) .or. any(av/=gv)) then
       write(std_err,*)'Locate module error',g,a
    else
       write(std_err,*)'Locate module is correct'
    end if
    ! now check with initial guesses
    a = lookup(p,x,(/2,3,2,3,2,2,3,3,2/))
    av = lookup(pv,x,y,z)
    if (any(a/=g) .or. any(av/=gv)) then
       write(std_err,*)'Locate module error',g,a
    else
       write(std_err,*)'Locate module is correct'
    end if
#endif
  end subroutine test_locate_module
  ! Finds which regions a piece of data is in

  ! allocates the passed in pointer
  subroutine find_data_s(r,lt,num)
    type(locate_t), intent(in) :: lt
    integer, intent(in) :: num
    type(rreg_t), pointer :: r(:)
    integer :: ARG(i,j,k),cnt 
    if (debug_locate) call write_locate_t(std_log,lt,'find data')
    cnt = 0
    do i=0,lt%nc%x
       do j=0,lt%nc%y
#ifdef THREEDIM
          do k=0,lt%nc%z 
#endif
             if (any(lt%d(ARG(i,j,k),1:lt%n(ARG(i,j,k)))==num)) then
                cnt = cnt+1
             end if
          CMD3(end do)
       end do
    end do
    allocate(r(cnt)) ! Allocate space for the result
    cnt = 1 
    do i=0,lt%nc%x
       do j=0,lt%nc%y
#ifdef THREEDIM
          do k=0,lt%nc%z 
#endif
             if (any(lt%d(ARG(i,j,k),1:lt%n(ARG(i,j,k)))==num)) then
                r(cnt) = get_region(lt,ivec_t(ARG(i,j,k)))
                cnt = cnt+1
             end if
          CMD3(end do)
       end do
    end do   
    if (debug_locate) write(std_log,*)'Find data finished' 
  end subroutine find_data_s   
  function get_region(lt,p) result(rg)
    type(locate_t), intent(in) :: lt
    type(ivec_t), intent(in) :: p
    type(rreg_t) rg
    if (p%x==0) then 
       rg%min%x = lt%rg%min%x
    else 
       rg%min%x = lt%xc(p%x)
    end if
    if (p%x==lt%nc%x) then 
       rg%max%x = lt%rg%max%x
    else 
       rg%max%x = lt%xc(p%x+1)
    end if
    if (p%y==0) then 
       rg%min%y = lt%rg%min%y
    else 
       rg%min%y = lt%yc(p%y)
    end if
    if (p%y==lt%nc%y) then 
       rg%max%y = lt%rg%max%y
    else 
       rg%max%y = lt%yc(p%y+1)
    end if
#ifdef THREEDIM
    if (p%z==0) then 
       rg%min%z = lt%rg%min%z
    else 
       rg%min%z = lt%zc(p%z)
    end if
    if (p%z==lt%nc%z) then 
       rg%max%z = lt%rg%max%z
    else 
       rg%max%z = lt%zc(p%z+1)
    end if
#endif
  end function get_region  
  subroutine ds_i_push_lt(d,lt,ARG(i,j,k))
    integer, intent(in) :: d
    type(locate_t), intent(inout) :: lt
    integer, intent(in) :: ARG(i,j,k)
    integer n
    n = lt%n(ARG(i,j,k))+1
    lt%n(ARG(i,j,k)) = n
    lt%d(ARG(i,j,k),n) = d
  end subroutine ds_i_push_lt
  subroutine da_i_push_lt(d,lt,ARG(i,j,k))
    integer, intent(in) :: d(:)
    type(locate_t), intent(inout) :: lt
    integer, intent(in) :: ARG(i,j,k)
    integer n
    n = lt%n(ARG(i,j,k))+size(d)
    lt%n(ARG(i,j,k)) = n
    lt%d(ARG(i,j,k),n-size(d)-1:n) = d
  end subroutine da_i_push_lt
end module locate
