! This module contains the routines for the particle dynamics
#include "part_props.h"
module pd 
  use precision
  use program_units
  use grid
  use vector
  use sysdep_io
  use particle
  use collision
  use wind
  implicit none
  !private SR2201BUG

  ! Public stuff  functions
  public :: particles_info     ! print information about the particles
  public :: pp_contact_info    ! print information about the particle conacts
  public :: particle_position_info ! PArticle position info
  public :: calc_pp_contacts   ! calculate particle-particle contacts
  public :: calc_pg_contacts   ! calculate particle-ghost contacts and forces
  public :: calc_pp_forces     ! calculate particle-particle forces
  public :: calc_displacements ! integrate forward in time
  public :: offset_velocities  ! Offset the velocities half a time step
  public :: check_particles    ! Check the particles aren't too close
  public :: init_forces        ! Set force to gravity and moment ot zero
  public :: calc_energy        ! Calculate potential and kinetic energy

  logical, private :: debug_pd = .false. ! Instigate extra checks
contains
    subroutine particles_info(part, fp)
      type(part_list_t), intent(in) :: part
      integer, intent(in) :: fp
      write(fp,*) 'Particle information' 
      if (associated(part%all))  then
         write(fp,*) 'Particle space allocated : ', size(part%all)
      else 
         write(fp,*) 'Particle space not allocated' 
      end if
      if (associated(part%p)) then
       write(fp,*) 'Particle space used : ', size(part%p) 
       write(fp,*) 'Grid square range is : ', minval(part%p%grid), maxval(part%p%grid)
       write(fp,*) 'Number of contacts, min, max, total : ', &
            minval(part%p%nc), maxval(part%p%nc), sum(part%p%nc)
       write(fp,*) 'Minimum coordinates : ', minval(part%p%mp%r)
       write(fp,*) 'Maximum coordinates : ', maxval(part%p%mp%r)
       write(fp,*) 'Minimum velocity : ', minval(part%p%mp%v)
       write(fp,*) 'Maximum velocity : ', maxval(part%p%mp%v)
       write(fp,*) 'Minimum angular velocity : ', minval(part%p%mp%w)
       write(fp,*) 'Maximum angular velocity : ', maxval(part%p%mp%w)
       write(fp,*) 'Minimum force : ', minval(part%p%f)
       write(fp,*) 'Maximum force : ', maxval(part%p%f)
       write(fp,*) 'Minimum angular moment : ', minval(part%p%f)
       write(fp,*) 'Maximum angular moment : ', maxval(part%p%f)
       write(fp,*) 'Contact potential energy min and max : ', &
             minval(part%p%cpe), maxval(part%p%cpe)
   else
       write(fp,*) 'Grid has not been allocated'
    end if
  end subroutine particles_info

  ! Print information about contacts
  subroutine pp_contact_info(part, fp)
    type(part_list_t), intent(in) :: part
    integer, intent(in) :: fp
    integer i;
    do i = 1,size(part%p)
       if (part%p(i)%nc==0 .and. size(part%p) > 10) cycle
       write(fp,*) 'particle ', i, ' is in contact with ', &
            part%p(i)%srn(1:part%p(i)%nc)
    end do
  end subroutine pp_contact_info
  subroutine particle_position_info(part, fp)
    type(part_list_t), intent(in) :: part
    integer, intent(in) :: fp
    integer i;
    write(fp,*) 'Particle position info'
    do i = 1,size(part%p)
       write(fp,*) 'Particle ', i, ' is at position ', part%p(i)%mp%r
    end do
  end subroutine particle_position_info
  subroutine init_forces(part)
     type(part_list_t), intent(inout) :: part
     if (debug_pd) write(std_log,*)"Initializing forces"
     part%p%f = P_MASS(part%p%mp)*gravity ! Initialise force to gravity
     part%p%am = null_rang ! Zero the force moment */
     part%p%cpe = 0 ! Zero the energy
   end subroutine init_forces
   
  ! Calculate particle particle contacts
  subroutine calc_pp_contacts(part)
    type(part_list_t), intent(inout) :: part
    integer n_con ! Number of contacts  
    integer i ! Count through the particles 
    integer j ! Other particle
    integer k ! count through particles in a grid square 
    real(r_kind) r ! separation distance */
    type(grid_element) ge ! Hold the particle numbers for this square
    type(gr_part_list) gpl ! Hold the list for all the surrounding particles

    do i = 1, size(part%p) ! Step through all the particles
       n_con = 0 ! Number of contacting particles
       ! look for new contacts with particles
       ! First in our own square with higher numbered particles
       call check_grid_point(ge, part%p(i)%grid)
       do k = 1, ge%n ! Look through all these particles
          j = ge%pn(k) ! Look at particle j
          if (i >= j) cycle ! only look at higher numbered particles
          r = distance_sqr(part%p(i)%mp%r,part%p(j)%mp%r)!Calc distance squared
          if (r < (P_RADIUS(part%p(i)%mp)+P_RADIUS(part%p(j)%mp))**2) then
             n_con = n_con+1
             part%p(i)%srn(n_con) = j  ! Contact with particle m
           end if
       end do
       ! Look through adjacent squares
       call h_adj_part(gpl, part%p(i)%grid)
       do k = 1, gpl%n ! Look through all these particles
          j = gpl%pn(k) ! Look at particle j
          r = distance_sqr(part%p(i)%mp%r,part%p(j)%mp%r)!Calc distance squared
          if (r < (P_RADIUS(part%p(i)%mp)+P_RADIUS(part%p(j)%mp))**2) then
             n_con = n_con+1
             part%p(i)%srn(n_con) = j  ! Contact with particle m
          end if
       end do
       part%p(i)%nc = n_con ! Number of contacts with   
    end do
  end subroutine calc_pp_contacts

  ! Calculate particle ghost contacts
  ! Easier than other routines as check contacts with any particles
  subroutine calc_pg_contacts(part,ghost)
    type(part_list_t), intent(inout) :: part
    type(min_part_list_t), intent(in) :: ghost
    integer i ! Count through the particles 
    integer j ! Other particle
    integer k ! count through particles in a grid square 
    real(r_kind) r ! separation distance */
     type(gr_part_list) gpl ! Hold the list for all the surrounding particles
    integer :: gc(size(ghost%p))     ! grid squares

    gc = coord_to_grid(ghost%p%r) ! Grid coordinates for ghost particles

    do i = 1, size(ghost%p) ! Step through all the particles
       call adj_part(gpl, gc(i))
       do k = 1, gpl%n ! Look through all these particles
          j = gpl%pn(k) ! Look at particle j
          r = distance_sqr(ghost%p(i)%r,part%p(j)%mp%r)!Calc distance squared
          if (r < (P_RADIUS(part%p(i)%mp)+P_RADIUS(part%p(j)%mp))**2) then
             call spring_force(part%p(j),ghost%p(i))
          end if
       end do
    end do
  end subroutine calc_pg_contacts

  subroutine calc_pp_forces(part)
    type(part_list_t), intent(inout) :: part
    integer i,j,k
    do i = 1, size(part%p)       ! Step through all the particles
       do k = 1, part%p(i)%nc  ! Step through all the contacts
          j = part%p(i)%srn(k)
          call spring_force(part%p(i),part%p(j))
       end do
    end do
  end subroutine calc_pp_forces

  subroutine calc_displacements(part)
    type(part_list_t), intent(inout) :: part
    part%p%mp%v = part%p%mp%v + (DT*P_IMASS(part%p%mp))*part%p%f  ! update linear velocity
    part%p%mp%w = part%p%mp%w + (DT*P_IMI(part%p%mp))*part%p%am   ! update angular velocity
    part%p%mp%r = part%p%mp%r + DT*part%p%mp%v       ! update position
  end subroutine calc_displacements
  ! perform half a time step backwards on the velocities
  ! since v and w are given at times (n-1/2)dt and
  subroutine offset_velocities(part)
    type(part_list_t), intent(inout) :: part
    part%p%mp%v = part%p%mp%v - (DT/2*P_IMASS(part%p%mp))*part%p%f  ! update linear velocity
    part%p%mp%w = part%p%mp%w - (DT/2*P_IMI(part%p%mp))*part%p%am   ! update angular velocity
  end subroutine offset_velocities

  ! Must be called before calc_displacemenst and after forces
  ! have been calculated
  ! The mass etc can be taken out of the sums if they're constant
  ! Do this using preprocessor if this is two slow
  subroutine calc_energy(part,energy) 
    type(part_list_t), intent(in) :: part
    type(energy_t), intent(out) :: energy
    ! Advance velocities to current time
    energy%lke =  sum(P_MASS(part%p%mp)/2*&
         length_sqr(part%p%mp%v + (DT*P_IMASS(part%p%mp)/2)*part%p%f))
    energy%ake = sum(P_MI(part%p%mp)/2 &
         *length_sqr(part%p%mp%w + (DT*P_IMI(part%p%mp)/2)*part%p%am))
    energy%gpe = -sum(P_MASS(part%p%mp)*part%p%mp%r .dot. gravity)! Gravitational p.e.
    energy%cpe = sum(part%p%cpe)        ! Contacts potential energy   
    energy%tke = energy%lke+energy%ake  ! Total kinetic energy 
    energy%tpe = energy%gpe+energy%cpe  ! Total potential energy 
    energy%total = energy%tke+energy%tpe ! Total energy 
   end subroutine calc_energy

  subroutine check_particles(part)
    type(part_list_t), intent(in) :: part
!!$    integer :: i,j,k
!!$    real(r_kind) :: min_sep
!!$    real(r_kind) :: vel_sqr
!!$    real(r_kind) :: d, closest, fastest
!!$    logical :: too_close, too_fast
!!$    too_close = .false.
!!$    too_fast = .false.
!!$    closest = big_real
!!$    fastest = 0
!!$    min_sep = col_props(PP_COLL)%max_ovlp
!!$    vel_sqr = col_props(PP_COLL)%max_vel**2
!!$    ! Check all the particles are too close
    call assert(isfinite(part%p),'Particles not finite')
!!$    do i = 1,size(part%p)
!!$       do j = 1, part%p(i)%nc
!!$          k = part%p(i)%srn(j)
!!$          d=distance_sqr(part%p(i)%mp%r,part%p(k)%mp%r)
!!$          if (d<(P_RADIUS(part%p(i))+P_RADIUS(part%p(j))-min_sep)**2) then
!!$             too_close = .true.
!!$             d = sqrt(d)-P_RADIUS(part%p(i))-P_RADIUS(part%p(j))
!!$             closest = min(closest,d)
!!$          end if
!!$          d = distance_sqr(part%p(i)%mp%v,part%p(k)%mp%v)
!!$          if (d > vel_sqr) then
!!$             too_fast = .true. 
!!$             fastest = max(fastest,sqrt(d))
!!$          end if
!!$         end do
!!$     end do
!!$     if (too_close) then
!!$          call conv_length_to_si(closest)
!!$        write(std_log,*)'Particles are too close',closest
!!$     end if
!!$     if (too_fast) then
!!$          call conv_velocity_to_si(fastest)
!!$        write(std_log,*)'Relative speed is to high',fastest
!!$     end if
  end subroutine check_particles
end module pd
