! This module is responsible for deciding when data should be recorded
! and what data should be recorded
! it is called at the beginning of a time step by all processors with the current
! time and step number (These only exist on the master processor however)
! Using this information it broadcasts a rec_flags_t structure to all
! processes with a true flag indicating that that data is to be stored this 
! time step
! these flags are made available globally by the rec_flags module
! so as other modules can avoid uneccasry computation 
 
! Definition for a derived type holdng information about what and when to record data
! this also includes an mpi interface for sending appropriate information to all 
! processes

! Thoughts on speeding things up.
! Permenantly allocate all record buffers at initialization time
! May mean a vary large memory allocation for root node

module rec_flags
  use rec_flags_f90_def
  type(rec_flags_t) rec_flag
end module rec_flags


! module for writing out program output
! reads a namelist which specifies which parameters to save and how often
module record 
  USE rec_flags  ! local module
  use precision
  use sysdep_io
  use vector
  use program_units
  use particle
  use min_part_f90_def
  use comms
  use rec_flags_f90_def
  use mpi
  implicit none
  !private SR2201BUG
 
  public :: read_record_param   ! Read the record parameters from a  namelist
  public :: start_record        ! Decide what to record this time step
  public :: make_record         ! record stuff for this time step     
  public :: end_record          ! Close all the record files
  public :: error_record        ! What to save if an error occurs
  public :: save_state          ! Save the whole state
  public :: rec_flag_energy     ! Are we recording energy this time step
  public :: save_part_props     ! Write all particle props 

  ! unit pointers for the various files
  integer,private ::  position_fp,energy_fp,velocity_fp,angular_fp,& 
       force_fp,moment_fp,stats_fp
  type(rec_flags_t),private :: base_rec
 
  real(r_kind),private :: rec_t = huge(rec_t)
  integer,private :: rec_n = 0                          ! Number of records written
  character(len=FILE_NAME_LEN),private  :: rec_name     ! base name for output files

  logical, parameter, private :: rec_debug = .false. ! Print extra messages
contains
  ! close all the files
  subroutine end_record
    call info_msg('Entering end_record',rec_debug)
    if (fmpi_is_root()) then  
       close(energy_fp)
       close(position_fp)
       close(velocity_fp)
       close(angular_fp)
       close(moment_fp)
       close(force_fp)
       close(stats_fp)
    end if
    call info_msg('Exiting end_record',rec_debug)
  end subroutine end_record
  ! If the program halts with an error stick everything out to this file
  subroutine error_record(p, fp)! 
    type(part_list_t), intent(in) :: p   ! particles
    integer, intent(in)           :: fp  ! program step
    call info_msg('Entering error_record',rec_debug)
    if (fmpi_is_root()) then
       energy_fp = fp
       position_fp = fp
       velocity_fp = fp
       angular_fp = fp
       moment_fp = fp
       force_fp = fp
       stats_fp = fp  
    end if
    rec_flag = all_rec  ! Set all record flags true
    call make_record(p, -1, -real(1,r_kind))
    call info_msg('Exiting error_record',rec_debug)
  end subroutine error_record
  ! check whtehr to make a record or not and broadcast the flags to all
  ! processes
  subroutine start_record(s ,t, force)
    integer, intent(in)           :: s   ! program step
    real(r_kind), intent(in)      :: t   ! program time
    logical, optional, intent(in) :: force ! Whether to force a record
    logical :: ff = .false.
    character(len=100) :: h_str
    if (present(force)) ff =  force
    if (fmpi_is_root()) then
       if (ff .or. t>=rec_n*rec_t) then
          rec_n=rec_n+1
          write(h_str,'(A5,I10,A5,F12.6)') "step", s,"time", t
          !       write(std_log,*)trim(h_str)
          call info_msg(h_str) 
          rec_flag = base_rec  ! Set the flags we want to record
       else
          rec_flag = null_rec  ! Set all the other flags to null
       end if
    end if
    call fmpi_bcast(rec_flag) ! Broadcast to all processes
  end subroutine start_record

  ! Collect all the necessary data onto the root node and save it
  !  The data tocollect and record is specified by the global rec_flag  
  subroutine make_record(p,s,t)   
    type(part_list_t), intent(in) :: p
    integer, intent(in)           :: s   ! program step
    real(r_kind), intent(in)      :: t   ! program time
    character(len=100) ::  h_str
    type(energy_t) e
    write(h_str,'(A5,I10,A5,F12.6)') "step", s,"time", t
    if (rec_flag%energy) then
 !      call calc_energy(p,e) ! Calculate the energy
       call conv_to_si(e)       ! 
       write(energy_fp,*)trim(h_str)
       write(energy_fp,*)e      
    end if
    if (rec_flag%position) then       
       call record_vec_data(h_str,p%p%mp%r,position_fp, L=1)
    end if
    if (rec_flag%velocity) then
       call record_vec_data(h_str,p%p%mp%v,velocity_fp, L=1,T=-1)
    end if
    if (rec_flag%angular) then
       call record_ang_data(h_str,p%p%mp%w,angular_fp, T=-1)
    end if
    if (rec_flag%force) then
       call record_vec_data(h_str,p%p%f,force_fp, M=1,L=1,T=-2)
    end if
    if (rec_flag%moment) then
       call record_ang_data(h_str,p%p%am,moment_fp, M=1,L=2,T=-2)
    end if
    if (rec_flag%stats) then  
       write(stats_fp,*)trim(h_str)
       call record_vec_stats_data('r',p%p%mp%r,stats_fp, L=1)
       call record_scalar_stats_data('v',length(p%p%mp%v),stats_fp,L=1,T=-1)
       call record_scalar_stats_data('w',length(p%p%mp%w),stats_fp,T=-1)
       call record_scalar_stats_data('f',length(p%p%f),stats_fp,M=1,L=1,T=-2)
       call record_scalar_stats_data('m',length(p%p%am),stats_fp,M=1,L=2,T=-2)
       write(stats_fp,*)'Total contacts',sum(p%p%nc)
    end if
  end subroutine make_record
  subroutine record_vec_data(s, v, fp,  T, L, M)
    character(len=*), intent(in)  :: s
    type(rvec_t), intent(in)       :: v(:)
    integer, intent(in)           :: fp    
    integer, intent(in), optional :: T,L,M
    type(rvec_t) sv(size(v))
    type(rvec_t), pointer       :: data(:)
    integer i
    sv = v
    call conv_to_si(sv,T,L,M)    ! Convert it to si on each node
    call fmpi_gathera(sv,data) ! Auto allocating and sizing mpi_gatherv 
    ! data is allocated to zero on all nodes except root
    if (fmpi_is_root()) then
       write(fp,*)trim(s)
       do i = 1, size(data)
          write(fp,*) i, data(i)
       end do
    end if
    deallocate(data)
  end subroutine record_vec_data
  subroutine record_ang_data(s, v, fp,  T, L, M)
    character(len=*), intent(in)  :: s
    type(rang_t), intent(in)       :: v(:)
    integer, intent(in)           :: fp    
    integer, intent(in), optional :: T,L,M
    type(rang_t) sv(size(v))
    type(rang_t), pointer       :: data(:)
    integer i
    sv = v
    call conv_to_si(sv,T,L,M)    ! Convert it to si on each node
    call fmpi_gathera(sv,data) ! Auto allocating and sizing mpi_gatherv 
    ! data is allocated to zero on all nodes except root
    if (fmpi_is_root()) then
       write(fp,*)trim(s)
       do i = 1, size(data)
          write(fp,*) i, data(i)
       end do
    end if
    deallocate(data)
  end subroutine record_ang_data
  subroutine record_scalar_stats_data(s,x,fp,T,L,M)
    character(len=*), intent(in)  :: s
    real(r_kind), intent(in)      :: x(:)
    integer, intent(in)           :: fp    
    integer, intent(in), optional :: T,L,M
    real(r_kind) :: r(3)
    integer n
    call conv_to_si(r,T,L,M)
    call fmpi_reduce(minval(x),r(1),fmpi_min)
    call fmpi_reduce(maxval(x),r(2),fmpi_min)
    call fmpi_reduce(sum(x),r(3),fmpi_sum)
    call fmpi_reduce(size(x),n,fmpi_sum)
    if (fmpi_is_root()) then
       r(3)=r(3)/n
       write(fp,*)trim(s),' min max mean',r
    end if
  end subroutine record_scalar_stats_data
  subroutine record_vec_stats_data(s,x,fp,T,L,M)
    character(len=*), intent(in)  :: s
    type(rvec_t), intent(in)         :: x(:)
    integer, intent(in)           :: fp    
    integer, intent(in), optional :: T,L,M
    type(rvec_t) :: r(3)
    integer n
    
    call conv_to_si(r,T,L,M)
    call fmpi_reduce(minval(x),r(1),fmpi_min)
    call fmpi_reduce(maxval(x),r(2),fmpi_max)
    call fmpi_reduce(sum(x),r(3),fmpi_sum)
    call fmpi_reduce(size(x),n,fmpi_sum)
    if (fmpi_is_root()) then
       r(3)=r(3)/n
       write(fp,*)trim(s),' min',r(1)
       write(fp,*)trim(s),' max',r(2)
       write(fp,*)trim(s),' mean',r(3)
    end if
  end subroutine record_vec_stats_data
  subroutine read_record_param(fp)
    integer, intent(in) :: fp
    logical :: rec_energy   = .false.  
    logical :: rec_velocity = .false. 
    logical :: rec_position = .false.  
    logical :: rec_angular  = .false.  
    logical :: rec_force    = .false.   
    logical :: rec_moment   = .false.   
    logical :: rec_stats    = .false.   
    namelist /record_param/ rec_energy,rec_velocity,rec_position,&
         rec_angular,rec_force,rec_moment,rec_stats,rec_t, rec_name

    read(fp,nml=record_param)     ! read what data we have to record
    base_rec%energy	= rec_energy   
    base_rec%velocity	= rec_velocity 
    base_rec%position	= rec_position 
    base_rec%angular	= rec_angular 
    base_rec%force  	= rec_force   
    base_rec%moment	= rec_moment  
    base_rec%stats	= rec_stats  

    rec_n = 0  
    if (rec_energy) then
       call assign_lun(energy_fp)
       open(energy_fp, file=trim(rec_name)//'.energy',&
            status='replace',action='write')
    end if
    if (rec_position) then
       call assign_lun(position_fp)
       open(position_fp, file=trim(rec_name)//'.position',&
            status='replace',action='write')
    end if
    if (rec_velocity) then
       call assign_lun(velocity_fp)
       open(velocity_fp, file=trim(rec_name)//'.velocity',&
            status='replace',action='write')
    end if
    if (rec_angular) then
       call assign_lun(angular_fp)
       open(angular_fp, file=trim(rec_name)//'.angular',&
            status='replace',action='write')
    end if
    if (rec_force) then
       call assign_lun(force_fp)
       open(force_fp, file=trim(rec_name)//'.force',&
            status='replace',action='write')
    end if
    if (rec_moment) then
       call assign_lun(moment_fp)
       open(moment_fp, file=trim(rec_name)//'.moment',&
            status='replace',action='write')
    end if
    if (rec_stats) then
       call assign_lun(stats_fp)
       open(stats_fp, file=trim(rec_name)//'.stats',&
            status='replace',action='write')
    end if
  end subroutine read_record_param 

  ! Save the data necessary to continue the simulation where we finished
  subroutine save_state(part)
    type(part_list_t), intent(in) :: part
    type(min_part_t) :: mp(size(part%p))
    type(min_part_t), pointer :: data(:)
    integer fp
    mp=part%p
    call conv_to_si(mp)
    call fmpi_gathera(mp,data) ! Auto allocating and sizing mpi_gatherv 

    if (fmpi_is_root()) then
       call assign_lun(fp)
       open(fp, file=trim(rec_name)//'.dump',status='replace',action='write')
       write(fp,*)data
!!$       call write_part_state(fp,data)
!!$       call write_part_props(fp,data)
       close(fp)
    end if
    deallocate(data)
  end subroutine save_state
  function rec_flag_energy()
    logical rec_flag_energy
    rec_flag_energy = rec_flag%energy
  end function rec_flag_energy
  subroutine save_part_props(p)
    type(min_part_t), intent(in) :: p(:)
    type(min_part_t) q(size(p))
    integer fp
    q = p
    call conv_to_si(q)
    call assign_lun(fp)
    open(fp, file=trim(rec_name)//'.pprops',status='replace',action='write')
    call write_part_props(fp,q)
    close(fp)
    call free_lun(fp)
  end subroutine save_part_props
end module record
