module wind
  use precision
  use sysdep_io
  use vector  
  use program_units
  use maths_const
  implicit none
  !private SR2201BUG
  public :: init_wind, calc_wind_forces, read_wind_param, wind_info
  public :: bcast_wind

  interface calc_wind_forces
     module procedure calc_wind_forces_sr, calc_wind_forces_si, &
          calc_wind_forces_vr
  end interface

  ! Parameters to be read in all in SI units
  real(r_kind),private :: air_density = 1.225       ! density Kg/m^3
  real(r_kind),private :: air_pressure = 0          ! pressure N/m^2
  real(r_kind),private :: air_viscosity = 0.0000145 ! Kinematic viscosity m^2/s
  real(r_kind),private :: air_temperature = 15      ! Degrees Centigrade 
  type(rvec_t),private :: air_velocity 
  real(r_kind),private :: terminal_velocity=7.5;
  logical,private      :: air_use
  namelist /air_param/ air_use, air_density, air_pressure, &
       air_viscosity ,air_temperature, air_velocity, terminal_velocity
  private :: air_param
  real(r_kind),private :: drag_param(2) 

contains
  subroutine bcast_wind
    use mpi
    call fmpi_bcast(air_use)
    call fmpi_bcast(drag_param)  ! Only these drag parameters to broadcast
  end subroutine bcast_wind
  subroutine init_wind
  end subroutine init_wind
  ! Read the wind parameters
  ! and calulate drag coefficents
  subroutine read_wind_param(fp,gravity)
    integer, intent(in) :: fp
    type(rvec_t),  intent(in) :: gravity
    real(r_kind) :: g
    real(r_kind) :: Cd, C1, C2 ! 1/s
    air_velocity = null_rvec !SR2201BUG
    read(fp, nml=air_param) ! Read in the parameters
    call conv_from_si(air_density,T=0,L=-3,M=1)
    call conv_from_si(air_pressure,T=-2,L=-1,M=1)
    call conv_from_si(air_viscosity,T=-1,L=2,M=0)
    call conv_velocity_from_si(terminal_velocity)
    g = length(gravity)
    ! Convert to program units
    C1 = 24*air_viscosity/2         
    C2 = 0.4              
    Cd = PI/(2*air_density)
    drag_param(1) = Cd*C1
    drag_param(2) = Cd*C2
  end subroutine read_wind_param
  function drag_function(speed) result(drag)
    real(r_kind), intent(in) :: speed(:)
    real(r_kind) :: drag(size(speed))
    drag = drag_param(1) + drag_param(2) * speed
  end function drag_function
  subroutine calc_wind_forces_vr(f, r, v) 
    type(rvec_t),intent(in) :: v(:)
    real(r_kind) :: r(size(v))
    type(rvec_t),intent(inout) :: f(size(v))
    type(rvec_t) :: wf(size(v))
    if (.not.air_use) return
    wf = air_velocity-v                     ! Calculate difference
    f = f + wf*r*(drag_param(1) + r*drag_param(2)*length(wf))
  end subroutine calc_wind_forces_vr
  subroutine calc_wind_forces_sr(f, r, v) 
    type(rvec_t),intent(in) :: v(:)
    real(r_kind) :: r
    type(rvec_t),intent(inout) :: f(size(v))
    type(rvec_t) :: wf(size(v))
    if (.not.air_use) return
    wf = air_velocity-v                     ! Calculate difference
    f = f + wf*(r*drag_param(1) + r**2*drag_param(2) * length(wf))
  end subroutine calc_wind_forces_sr
  subroutine calc_wind_forces_si(f, r, v) 
    type(rvec_t),intent(in) :: v(:)
    integer :: r
    type(rvec_t),intent(inout) :: f(size(v))
    type(rvec_t) :: wf(size(v))
    if (.not.air_use) return
    wf = air_velocity-v                     ! Calculate difference
    f = f + wf*(r*drag_param(1) + r**2*drag_param(2) * length(wf))
  end subroutine calc_wind_forces_si
  subroutine wind_info(fp)
      integer, intent(in) :: fp
      write(fp,*) 'Wind module information' 
      write(fp,*) 'air density', air_density
      write(fp,*) 'air pressure', air_pressure
      write(fp,*) 'air viscosity', air_viscosity
      write(fp,*) 'air velocity', air_velocity
      call conv_velocity_to_si(terminal_velocity)
      write(fp,*) 'SI terminal velocity', terminal_velocity
      call conv_velocity_from_si(terminal_velocity)
      write(fp,*) 'PU terminal velocity', terminal_velocity
      write(fp,*) 'drag_param', drag_param
      if (air_use) then 
         write(fp,*) 'Using air module'
      else
        write(fp,*) 'Not using air module'
     end if
  end subroutine wind_info
end module
